"""
MikoshiLang Extended Functions - Set 25
Materials science, nanotechnology, and polymer science
"""
import sympy as sp
from sympy import symbols, sqrt, exp, log, pi, sin, cos

def get_rules():
    """Extended set 25: Materials and nanoscience (150 functions)"""
    T, E, sigma = symbols('T E sigma')
    
    return [
        # ===== CRYSTALLOGRAPHY (30 functions) =====
        ('BravaisLattice[type_]', lambda t: str(t)),
        ('CubicLatticeParameter[a_]', lambda a: a),
        ('TetragonalLatticeParameter[a_, c_]', lambda a, c: (a, c)),
        ('OrthorhombicLatticeParameter[a_, b_, c_]', lambda a, b, c: (a, b, c)),
        ('HexagonalLatticeParameter[a_, c_]', lambda a, c: (a, c)),
        ('RhombohedralLatticeParameter[a_, alpha_]', lambda a, alpha: (a, alpha)),
        ('MonoclinicLatticeParameter[a_, b_, c_, beta_]', lambda a, b, c, beta: (a, b, c, beta)),
        ('TriclinicLatticeParameter[a_, b_, c_, alpha_, beta_, gamma_]', lambda a, b, c, alpha, beta, gamma: (a, b, c, alpha, beta, gamma)),
        ('ReciprocalLatticeVector[a_, b_, c_]', lambda a, b, c: 2*pi/a),
        ('ZoneAxis[h_, k_, l_]', lambda h, k, l: (h, k, l)),
        ('XRayDiffraction[n_, lambda_, d_, theta_]', lambda n, lam, d, theta: n*lam - 2*d*sin(theta)),
        ('BraggAngle[n_, lambda_, d_]', lambda n, lam, d: sp.asin(n*lam/(2*d))),
        ('StructureFactor[f_, phase_]', lambda f, phi: f*exp(sp.I*phi)),
        ('AtomicScatteringFactor[Z_, sin_theta_lambda_]', lambda Z, stl: Z*exp(-sp.Symbol('B')*stl**2)),
        ('LaueEquation[k_in_, k_out_, G_]', lambda kin, kout, G: kout - kin - G),
        ('EwaldSphere[k_]', lambda k: 2*pi/sp.Symbol('lambda')),
        ('PowderDiffraction[d_, theta_]', lambda d, theta: 2*d*sin(theta)),
        ('TextureCoefficient[I_hkl_, I0_hkl_]', lambda I, I0: I/I0),
        ('CrystalliteSize[K_, lambda_, beta_, theta_]', lambda K, lam, beta, theta: K*lam/(beta*cos(theta))),  # Scherrer
        ('MicrostrainBroadening[epsilon_, theta_]', lambda eps, theta: 4*eps*tan(theta)),
        ('WilliamsonHallPlot[beta_, theta_]', lambda beta, theta: beta*cos(theta)),
        ('DefectDensity[broadening_]', lambda b: b**2),
        ('StickingCoefficient[adsorbed_, incident_]', lambda ads, inc: ads/inc),
        ('SurfaceEnergy[gamma_]', lambda gamma: gamma),
        ('GrainBoundaryEnergy[theta_]', lambda theta: sp.Symbol('gamma_gb')*theta*(1 - log(theta))),
        ('TwinBoundaryEnergy[gamma_]', lambda gamma: gamma/2),
        ('InterfacialEnergy[gamma1_, gamma2_, gamma12_]', lambda g1, g2, g12: g1 + g2 - g12),
        ('WettingAngle[gamma_sv_, gamma_sl_, gamma_lv_]', lambda gsv, gsl, glv: sp.acos((gsv - gsl)/glv)),
        ('YoungDupreEquation[gamma_sv_, gamma_sl_, gamma_lv_]', lambda gsv, gsl, glv: gsv - gsl - glv*cos(sp.Symbol('theta'))),
        ('ContactAngleMeasurement[theta_]', lambda theta: theta),
        
        # ===== PHASE TRANSFORMATIONS (30 functions) =====
        ('NucleationRate[n0_, delta_G_]', lambda n0, dG: n0*exp(-dG/(sp.Symbol('k_B')*sp.Symbol('T')))),
        ('CriticalNucleusRadius[sigma_, delta_G_v_]', lambda sig, dGv: -2*sig/dGv),
        ('NucleationBarrier[sigma_, delta_G_v_]', lambda sig, dGv: 16*pi*sig**3/(3*dGv**2)),
        ('HomogeneousNucleation[I0_, delta_G_star_]', lambda I0, dGstar: I0*exp(-dGstar/(sp.Symbol('k_B')*sp.Symbol('T')))),
        ('HeterogeneousNucleation[f_theta_]', lambda f: f),  # Shape factor
        ('GrowthRate[mobility_, driving_force_]', lambda mu, F: mu*F),
        ('DiffusionControlledGrowth[D_, C_, t_]', lambda D, C, t: sqrt(D*t)),
        ('InterfaceControlledGrowth[v_, delta_G_]', lambda v, dG: v*dG),
        ('JohnsonMehlAvramiEquation[t_, k_, n_]', lambda t, k, n: 1 - exp(-k*t**n)),
        ('AvramiExponent[mechanism_]', lambda m: 3 if str(m) == '3D' else 2),
        ('KolmogorovModel[V_, t_]', lambda V, t: V*(1 - exp(-sp.Symbol('k')*t**sp.Symbol('n')))),
        ('TTTCurve[T_, time_]', lambda T, t: (T, log(t))),
        ('CCTCurve[T_, cooling_rate_]', lambda T, cr: (T, cr)),
        ('MartensiticTransformation[Ms_, Mf_]', lambda Ms, Mf: (Ms, Mf)),
        ('MartensiteStartTemperature[composition_]', lambda comp: 500),  # Simplified
        ('MartensiteFinishTemperature[Ms_]', lambda Ms: Ms - 200),
        ('BainiteTTransformation[T_]', lambda T: T),
        ('PearliteFormation[T_]', lambda T: T),
        ('OstwardRipening[r_, t_]', lambda r, t: r**3 - sp.Symbol('r0')**3 == sp.Symbol('k')*t),  # LSW theory
        ('CoarseningRate[gamma_, D_]', lambda gamma, D: gamma*D),
        ('SpinoidalDecomposition[c_, c_crit_]', lambda c, cc: c != cc),
        ('ChemicalSpinodal[d2G_dc2_]', lambda d2G: d2G < 0),
        ('CoherentSpinodal[elastic_, chemical_]', lambda el, ch: el + ch < 0),
        ('CahnHilliardEquation[c_, t_]', lambda c, t: sp.diff(c, t)),
        ('UpHillDiffusion[flux_]', lambda J: J > 0),  # Against concentration gradient
        ('PhaseFieldModel[phi_, eta_]', lambda phi, eta: (phi, eta)),
        ('OrderParameter[psi_]', lambda psi: psi),
        ('GinzburgLandauEquation[psi_]', lambda psi: sp.diff(sp.Symbol('F'), psi)),
        ('EutecticSolidification[T_]', lambda T: T),
        ('PeritecticSolidification[T_]', lambda T: T),
        
        # ===== MECHANICAL PROPERTIES (30 functions) =====
        ('TruStress[F_, A_]', lambda F, A: F/A),
        ('TrueStrain[l_, l0_]', lambda l, l0: log(l/l0)),
        ('EngineeringStress[F_, A0_]', lambda F, A0: F/A0),
        ('EngineeringStrain[l_, l0_]', lambda l, l0: (l - l0)/l0),
        ('StrainHardeningExponent[n_]', lambda n: n),
        ('HollomanEquation[K_, epsilon_, n_]', lambda K, eps, n: K*eps**n),
        ('ConsidereCriterion[sigma_, epsilon_]', lambda sig, eps: sp.diff(sigma, eps) == sigma/eps),
        ('NeckingInstability[d_sigma_d_epsilon_]', lambda dsigdeps: dsigdeps == sp.Symbol('sigma')/sp.Symbol('epsilon')),
        ('PlasticDeformation[epsilon_p_]', lambda epsp: epsp),
        ('ElasticRecovery[epsilon_e_]', lambda epse: epse),
        ('WorkHardening[d_sigma_d_epsilon_]', lambda dsigdeps: dsigdeps),
        ('RecoveryProcess[rho_]', lambda rho: -sp.diff(rho, sp.Symbol('t'))),
        ('RecrystallizationKinetics[f_, t_]', lambda f, t: 1 - exp(-sp.Symbol('k')*t**sp.Symbol('n'))),
        ('GrainGrowthKinetics[D_, t_]', lambda D, t: D**2 - sp.Symbol('D0')**2 == sp.Symbol('k')*t),
        ('HallPetchSlope[k_HP_]', lambda k: k),
        ('PetchRelation[sigma_y_, sigma_0_, k_, d_]', lambda sigy, sig0, k, d: sig0 + k/sqrt(d)),
        ('DislocationVelocity[stress_, b_]', lambda sig, b: sp.Symbol('v0')*(sig/sp.Symbol('sigma0'))**sp.Symbol('m')),
        ('OrowanimechansimBypass[tau_, G_, b_, r_]', lambda tau, G, b, r: G*b/r),
        ('TaylorHardening[alpha_, G_, b_, rho_]', lambda alpha, G, b, rho: alpha*G*b*sqrt(rho)),
        ('BaileyNortonCreep[sigma_, T_, t_]', lambda sig, T, t: sp.Symbol('A')*sig**sp.Symbol('n')*t**sp.Symbol('m')*exp(-sp.Symbol('Q')/(sp.Symbol('R')*T))),
        ('PowerLawCreep[epsilon_dot_, sigma_]', lambda epsdot, sig: epsdot == sp.Symbol('A')*sig**sp.Symbol('n')*exp(-sp.Symbol('Q')/(sp.Symbol('R')*sp.Symbol('T')))),
        ('DiffusionalCreep[sigma_, d_]', lambda sig, d: sp.Symbol('A')*sig/d**2),  # Nabarro-Herring
        ('GrainBoundarySliding[epsilon_, m_]', lambda eps, m: eps**m),
        ('SuperplasticDeformation[epsilon_dot_, sigma_, d_]', lambda epsdot, sig, d: epsdot == sp.Symbol('A')*sig**2/d**2),
        ('LarsonMillerParameter[T_, t_]', lambda T, t: T*(20 + log(t))),
        ('MansonHaferdParameter[T_, t_]', lambda T, t: (T - sp.Symbol('Ta'))/(log(t) - log(sp.Symbol('ta')))),
        ('MonkmanGrantRelation[epsilon_dot_, t_r_]', lambda epsdot, tr: epsdot*tr == sp.Symbol('C')),
        ('RobinsonLifeFraction[t_, t_r_]', lambda t, tr: sum(ti/tri for ti, tri in zip(t, tr))),
        ('CreepdamageParameter[omega_]', lambda omega: omega),
        ('KachanovDamage[sigma_, omega_]', lambda sig, omega: sig/(1 - omega)),
        
        # ===== POLYMERS (30 functions) =====
        ('DegreeOfPolymerization[n_]', lambda n: n),
        ('NumberAverageMolecularWeight[Mn_]', lambda Mn: Mn),
        ('WeightAverageMolecularWeight[Mw_]', lambda Mw: Mw),
        ('PolydispersityIndex[Mw_, Mn_]', lambda Mw, Mn: Mw/Mn),
        ('EndToEndDistance[N_, l_]', lambda N, l: sqrt(N)*l),
        ('RadiusOfGyration[Rg_]', lambda Rg: Rg),
        ('FloryExponent[nu_]', lambda nu: nu),  # 0.5 ideal, 0.6 good solvent
        ('MarkHouwinkEquation[M_, K_, a_]', lambda M, K, a: K*M**a),  # Intrinsic viscosity
        ('FloryHugginsParameter[chi_]', lambda chi: chi),
        ('ThetaTemperature[T_]', lambda T: T),  # Ideal polymer solution
        ('GlassTransitionTemperature[Tg_]', lambda Tg: Tg),
        ('MeltingPointPolymer[Tm_]', lambda Tm: Tm),
        ('CrystallinityDegree[Xc_]', lambda Xc: Xc),
        ('RubberElasticity[sigma_, lambda_]', lambda sig, lam: sig == sp.Symbol('G')*(lam - 1/lam**2)),
        ('EntropyElasticity[S_, lambda_]', lambda S, lam: -sp.diff(S, lam)),
        ('FloryRehnerEquation[V2_, chi_]', lambda V2, chi: log(1 - V2) + V2 + chi*V2**2),
        ('WilliamsLandelFerryEquation[T_, Tg_]', lambda T, Tg: -sp.Symbol('C1')*(T - Tg)/(sp.Symbol('C2') + T - Tg)),
        ('TimeTemperatureSuperposition[aT_]', lambda aT: aT),
        ('MaxwellModel[G_, tau_]', lambda G, tau: G*exp(-sp.Symbol('t')/tau)),
        ('KelvinVoigtModel[E_, eta_, t_]', lambda E, eta, t: E*(1 - exp(-E*t/eta))),
        ('StandardLinearSolid[E1_, E2_, eta_]', lambda E1, E2, eta: E1 + E2*(1 - exp(-sp.Symbol('t')*E2/eta))),
        ('ViscoelasticCompliance[J_]', lambda J: J),
        ('RelaxationModulus[G_, t_]', lambda G, t: G*exp(-t/sp.Symbol('tau'))),
        ('RetardationTime[tau_]', lambda tau: tau),
        ('LossModulus[G_double_prime_]', lambda Gpp: Gpp),
        ('StorageModulus[G_prime_]', lambda Gp: Gp),
        ('LossTangent[G_double_prime_, G_prime_]', lambda Gpp, Gp: Gpp/Gp),
        ('DynamicMechanicalAnalysis[E_star_]', lambda Estar: (sp.re(Estar), sp.im(Estar))),
        ('ThixotropicBehavior[viscosity_, time_]', lambda eta, t: eta*exp(-sp.Symbol('k')*t)),
        ('RheopexyBehavior[viscosity_, time_]', lambda eta, t: eta*(1 + sp.Symbol('k')*t)),
        
        # ===== NANOMATERIALS (30 functions) =====
        ('QuantumConfinement[E_g_, d_]', lambda Eg, d: Eg + sp.Symbol('C')/d**2),
        ('QuantumDotEnergy[n_, L_]', lambda n, L: n**2*sp.Symbol('h')**2/(8*sp.Symbol('m')*L**2)),
        ('ExcitonBohrRadius[epsilon_, m_]', lambda eps, m: eps*sp.Symbol('a0')*sp.Symbol('me')/m),
        ('SurfaceToVolumeRatio[d_]', lambda d: 6/d),  # For sphere
        ('NanoparticleMeltingPoint[Tm_bulk_, d_]', lambda Tm, d: Tm*(1 - sp.Symbol('alpha')/d)),
        ('GibbsThomsonEffect[r_]', lambda r: sp.Symbol('delta_T') == 2*sp.Symbol('gamma')*sp.Symbol('Tm')/(sp.Symbol('L')*sp.Symbol('rho')*r)),
        ('CarbonNanotubeBandGap[n_, m_]', lambda n, m: 0 if (n - m) % 3 == 0 else sp.Symbol('Eg')),
        ('GrapheneDispersion[k_]', lambda k: sp.Symbol('hbar')*sp.Symbol('v_F')*k),
        ('PlasmonResonance[omega_p_]', lambda omegap: omegap),
        ('LocalizedSurfacePlasmon[epsilon_, R_]', lambda eps, R: sp.Symbol('omega_p')/sqrt(1 + 2*eps)),
        ('MieScatteringNano[r_, lambda_]', lambda r, lam: (2*pi*r/lam)**4),
        ('RayleighScatteringNano[lambda_]', lambda lam: lam**(-4)),
        ('QuantumYield[photons_emitted_, photons_absorbed_]', lambda emit, abs: emit/abs),
        ('StokesShift[lambda_abs_, lambda_em_]', lambda labdabs, lambem: lambem - labdabs),
        ('FluorescenceLifetime[tau_]', lambda tau: tau),
        ('PhotoluminescenceQuantumEfficiency[radiative_, total_]', lambda rad, tot: rad/tot),
        ('SuperparamagnetismThreshold[K_, V_, T_]', lambda K, V, T: K*V/(sp.Symbol('k_B')*T)),
        ('BlockingTemperature[K_, V_]', lambda K, V: K*V/sp.Symbol('k_B')),
        ('CoercivityNanoparticle[K_, Ms_]', lambda K, Ms: 2*K/Ms),
        ('SaturationMagnetization[M_bulk_, d_]', lambda M, d: M*(1 - sp.Symbol('delta')/d)),
        ('NanoporesConductance[d_, l_]', lambda d, l: sp.Symbol('sigma')*pi*d**2/(4*l)),
        ('CapillaryCondensation[r_, gamma_, V_m_]', lambda r, gamma, Vm: sp.Symbol('p')/sp.Symbol('p0') == exp(-2*gamma*Vm/(sp.Symbol('R')*sp.Symbol('T')*r))),
        ('BETSurfaceArea[V_m_, N_A_, a_m_]', lambda Vm, NA, am: Vm*NA*am),
        ('LangmuirIsotherm[theta_, K_, p_]', lambda theta, K, p: K*p/(1 + K*p)),
        ('FreundlichIsotherm[q_, K_, p_, n_]', lambda q, K, p, n: K*p**sp.Rational(1,n)),
        ('DLVOTheory[V_A_, V_R_]', lambda VA, VR: VA + VR),
        ('vanderWaalsAttraction[A_, d_]', lambda A, d: -A/(12*pi*d**2)),
        ('ElectrostaticRepulsion[psi_, kappa_, d_]', lambda psi, kappa, d: 64*sp.Symbol('n')*sp.Symbol('k_B')*sp.Symbol('T')*sp.tanh(sp.Symbol('e')*psi/(4*sp.Symbol('k_B')*sp.Symbol('T')))**2*exp(-kappa*d)),
        ('DebyeLength[epsilon_, T_, c_]', lambda eps, T, c: sqrt(eps*sp.Symbol('k_B')*T/(2*sp.Symbol('e')**2*sp.Symbol('N_A')*c))),
        ('ZetaPotential[mu_, epsilon_, eta_]', lambda mu, eps, eta: mu*eta/eps),
    ]


def register():
    """Register all extended25 rules"""
    from .extended_helper import convert_rules
    return convert_rules(get_rules())



"""Built-in functions — trig, exponential, number theory, combinatorics, logic."""

from __future__ import annotations
import math
import sympy as sp
from .expr import Expr, Symbol
from .math_ops import to_sympy, from_sympy


# ── Expression constructors (symbolic) ──

def Sin(x) -> Expr:
    return Expr("Sin", x)

def Cos(x) -> Expr:
    return Expr("Cos", x)

def Tan(x) -> Expr:
    return Expr("Tan", x)

def ArcSin(x) -> Expr:
    return Expr("ArcSin", x)

def ArcCos(x) -> Expr:
    return Expr("ArcCos", x)

def ArcTan(x) -> Expr:
    return Expr("ArcTan", x)

def Exp(x) -> Expr:
    return Expr("Exp", x)

def Log(x, base=None) -> Expr:
    if base is not None:
        return Expr("Log", base, x)
    return Expr("Log", x)

def Log2(x) -> Expr:
    return Expr("Log", 2, x)

def Log10(x) -> Expr:
    return Expr("Log", 10, x)


# ── Number theory ──

def Prime(n: int) -> int:
    """Return the nth prime number."""
    return int(sp.prime(n))

def PrimeQ(n: int) -> bool:
    """Test if n is prime."""
    return sp.isprime(n)

def FactorInteger(n: int) -> Expr:
    """Factor an integer into prime factors."""
    factors = sp.factorint(n)
    return Expr("List", *(Expr("List", int(p), int(e)) for p, e in sorted(factors.items())))

def GCD(*args: int) -> int:
    """Greatest common divisor."""
    result = args[0]
    for a in args[1:]:
        result = math.gcd(result, a)
    return result

def LCM(*args: int) -> int:
    """Least common multiple."""
    result = args[0]
    for a in args[1:]:
        result = result * a // math.gcd(result, a)
    return result

def Mod(a: int, b: int) -> int:
    """Modular arithmetic."""
    return a % b


# ── Combinatorics ──

def Factorial(n: int) -> int:
    """n!"""
    return math.factorial(n)

def Binomial(n: int, k: int) -> int:
    """Binomial coefficient C(n, k)."""
    return math.comb(n, k)

def Fibonacci(n: int) -> int:
    """nth Fibonacci number."""
    return int(sp.fibonacci(n))


# ── Logic ──

def And(*args: bool) -> bool:
    return all(args)

def Or(*args: bool) -> bool:
    return any(args)

def Not(x: bool) -> bool:
    return not x

def Xor(a: bool, b: bool) -> bool:
    return a ^ b

def Implies(a: bool, b: bool) -> bool:
    return (not a) or b


# ── Comparison ──

def Equal(a, b) -> bool:
    return a == b

def Less(a, b) -> bool:
    return a < b

def Greater(a, b) -> bool:
    return a > b

def LessEqual(a, b) -> bool:
    return a <= b

def GreaterEqual(a, b) -> bool:
    return a >= b


# ── Registry ──

BUILTIN_FUNCTIONS = {
    "Sin": Sin, "Cos": Cos, "Tan": Tan,
    "ArcSin": ArcSin, "ArcCos": ArcCos, "ArcTan": ArcTan,
    "Exp": Exp, "Log": Log, "Log2": Log2, "Log10": Log10,
    "Prime": Prime, "PrimeQ": PrimeQ, "FactorInteger": FactorInteger,
    "GCD": GCD, "LCM": LCM, "Mod": Mod,
    "Factorial": Factorial, "Binomial": Binomial, "Fibonacci": Fibonacci,
    "And": And, "Or": Or, "Not": Not, "Xor": Xor, "Implies": Implies,
    "Equal": Equal, "Less": Less, "Greater": Greater,
    "LessEqual": LessEqual, "GreaterEqual": GreaterEqual,
}



"""
MikoshiLang Extended Functions - Set 36
Manufacturing, production systems, and industrial engineering
"""
import sympy as sp
from sympy import symbols, sqrt, exp, log, pi

def get_rules():
    """Extended set 36: Manufacturing and industrial engineering (150 functions)"""
    x, t, n = symbols('x t n')
    
    return [
        # ===== PRODUCTION PLANNING (35 functions) =====
        ('ProductionRate[output_, time_]', lambda out, t: out/t),
        ('CycleTime[total_time_, units_]', lambda tt, units: tt/units),
        ('TaktTime[available_time_, demand_]', lambda at, d: at/d),
        ('LeadTime[processing_, waiting_, transport_]', lambda proc, wait, trans: proc + wait + trans),
        ('ThroughputTime[start_, end_]', lambda start, end: end - start),
        ('UtilizationRate[actual_, available_]', lambda actual, avail: actual/avail*100),
        ('EfficiencyRate[actual_, standard_]', lambda actual, std: actual/std*100),
        ('OverallEquipmentEffectiveness[availability_, performance_, quality_]', lambda A, P, Q: A*P*Q),
        ('AvailabilityFactor[uptime_, total_time_]', lambda up, tot: up/tot*100),
        ('PerformanceRate[actual_speed_, ideal_speed_]', lambda actual, ideal: actual/ideal*100),
        ('QualityRate[good_units_, total_units_]', lambda good, total: good/total*100),
        ('CapacityUtilization[actual_, capacity_]', lambda actual, cap: actual/cap*100),
        ('BottleneckIdentification[station_times_]', lambda times: max(times)),
        ('LineBalancing[task_times_, num_stations_]', lambda tt, n: sum(tt)/n),
        ('BalanceDelay[idle_time_, total_time_]', lambda idle, total: idle/total*100),
        ('SmoothingIndex[variation_, mean_]', lambda var, mean: var/mean),
        ('ProductionSmoothing[actual_, target_]', lambda actual, target: abs(actual - target)),
        ('MasterProductionSchedule[demand_, inventory_]', lambda d, inv: d - inv),
        ('MaterialRequirementsPlanning[gross_, scheduled_, on_hand_]', lambda gross, sched, oh: gross - sched - oh),
        ('SafetyStock[demand_variability_, lead_time_]', lambda dv, lt: dv*sqrt(lt)),
        ('ReorderPoint[demand_rate_, lead_time_, safety_stock_]', lambda dr, lt, ss: dr*lt + ss),
        ('EconomicOrderQuantity[demand_, ordering_cost_, holding_cost_]', lambda D, S, H: sqrt(2*D*S/H)),
        ('OrderingCost[orders_, cost_per_order_]', lambda orders, cpo: orders*cpo),
        ('HoldingCost[inventory_, holding_rate_]', lambda inv, hr: inv*hr),
        ('TotalInventoryCost[ordering_, holding_]', lambda ord, hold: ord + hold),
        ('InventoryTurnoverRatio[COGS_, avg_inventory_]', lambda cogs, inv: cogs/inv),
        ('DaysInventoryOutstanding[inventory_, daily_COGS_]', lambda inv, dcogs: inv/dcogs),
        ('StockoutCost[units_short_, cost_per_unit_]', lambda short, cpu: short*cpu),
        ('ServiceLevel[demand_met_, total_demand_]', lambda met, total: met/total*100),
        ('FillRate[lines_filled_, total_lines_]', lambda filled, total: filled/total*100),
        ('BackorderLevel[demand_, supply_]', lambda d, s: max(0, d - s)),
        ('LotSizing[setup_cost_, holding_cost_, demand_]', lambda setup, hold, d: sqrt(2*setup*d/hold)),
        ('CapacityRequirementPlanning[orders_, capacity_]', lambda orders, cap: orders/cap),
        ('LoadingFactor[load_, capacity_]', lambda load, cap: load/cap*100),
        ('SchedulingPriority[due_date_, processing_time_]', lambda dd, pt: dd - pt),
        
        # ===== QUALITY CONTROL (30 functions) =====
        ('DefectRate[defects_, total_]', lambda def_count, total: def_count/total*100),
        ('YieldRate[good_, total_]', lambda good, total: good/total*100),
        ('ScrapRate[scrap_, total_]', lambda scrap, total: scrap/total*100),
        ('ReworkRate[rework_, total_]', lambda rework, total: rework/total*100),
        ('FirstPassYield[good_first_, total_]', lambda good, total: good/total*100),
        ('RolledThroughputYield[FPY_stages_]', lambda fpys: sp.prod(fpys)),
        ('SigmaLevel[defects_per_million_]', lambda dpm: 6 - sp.log(dpm/1e6, 2.718)/sqrt(2)),
        ('ProcessCapability[tolerance_, process_variation_]', lambda tol, var: tol/(6*var)),
        ('CpIndex[USL_, LSL_, sigma_]', lambda USL, LSL, sigma: (USL - LSL)/(6*sigma)),
        ('CpkIndex[mean_, USL_, LSL_, sigma_]', lambda mean, USL, LSL, sigma: min((USL - mean)/(3*sigma), (mean - LSL)/(3*sigma))),
        ('PpmDefectRate[defects_, total_]', lambda def_count, total: def_count/total*1e6),
        ('ControlChartUCL[mean_, sigma_, n_]', lambda mean, sigma, n: mean + 3*sigma/sqrt(n)),
        ('ControlChartLCL[mean_, sigma_, n_]', lambda mean, sigma, n: mean - 3*sigma/sqrt(n)),
        ('ControlChartCenterLine[mean_]', lambda mean: mean),
        ('RChart Range[max_, min_]', lambda max_val, min_val: max_val - min_val),
        ('XBarChart[sample_mean_]', lambda xbar: xbar),
        ('PChart[defectives_, sample_size_]', lambda d, n: d/n),
        ('CChart[defects_]', lambda c: c),
        ('SampleSize[Z_, p_, E_]', lambda Z, p, E: (Z**2*p*(1 - p))/E**2),
        ('ConfidenceInterval[mean_, Z_, sigma_, n_]', lambda mean, Z, sigma, n: (mean - Z*sigma/sqrt(n), mean + Z*sigma/sqrt(n))),
        ('TypeIError[alpha_]', lambda alpha: alpha),
        ('TypeIIError[beta_]', lambda beta: beta),
        ('PowerOfTest[1_minus_beta_]', lambda beta: 1 - beta),
        ('AcceptanceSampling[n_, c_]', lambda n, c: (n, c)),
        ('AQL[acceptable_quality_level_]', lambda aql: aql),
        ('LTPD[lot_tolerance_percent_defective_]', lambda ltpd: ltpd),
        ('ProducerRisk[alpha_]', lambda alpha: alpha),
        ('ConsumerRisk[beta_]', lambda beta: beta),
        ('OperatingCharacteristicCurve[p_, n_, c_]', lambda p, n, c: sp.Symbol('Pa')),
        ('TaguchLossFunction[y_, target_, k_]', lambda y, T, k: k*(y - T)**2),
        
        # ===== LEAN MANUFACTURING (25 functions) =====
        ('ValueAddedTime[processing_time_]', lambda pt: pt),
        ('NonValueAddedTime[waiting_, transport_, inspection_]', lambda wait, trans, insp: wait + trans + insp),
        ('ValueAddedRatio[VA_time_, total_time_]', lambda va, total: va/total*100),
        ('ProcessCycleEfficiency[VA_time_, lead_time_]', lambda va, lt: va/lt*100),
        ('WorkInProgress[cycle_time_, throughput_]', lambda ct, tp: ct*tp),
        ('LittlesLaw[WIP_, throughput_]', lambda wip, tp: wip/tp),
        ('FlowTime[WIP_, arrival_rate_]', lambda wip, arr: wip/arr),
        ('WaitingTime[queue_length_, service_rate_]', lambda q, mu: q/mu),
        ('QueueLength[arrival_rate_, service_rate_]', lambda lam, mu: lam/(mu - lam)),
        ('ServerUtilization[arrival_rate_, service_rate_]', lambda lam, mu: lam/mu),
        ('SetupReduction[current_, improved_]', lambda curr, imp: (curr - imp)/curr*100),
        ('ChangeoverTime[setup_, adjust_, trial_]', lambda setup, adj, trial: setup + adj + trial),
        ('SMED[single_minute_exchange_]', lambda smed: smed < 10),
        ('PokayokeEffectiveness[defects_prevented_, total_defects_]', lambda prev, total: prev/total*100),
        ('KanbanCards[demand_, lead_time_, container_size_]', lambda d, lt, cs: d*lt/cs),
        ('PullSystemSize[daily_demand_, lead_time_days_]', lambda dd, ltd: dd*ltd),
        ('Supermarket Size[max_inventory_]', lambda max_inv: max_inv),
        ('KaizenEvents[improvements_]', lambda imp: imp),
        ('ValueStreamMapping[activities_]', lambda act: act),
        ('FlowEfficiency[processing_, total_]', lambda proc, total: proc/total*100),
        ('TransportationWaste[distance_, trips_]', lambda dist, trips: dist*trips),
        ('MotionWaste[unnecessary_movements_]', lambda moves: moves),
        ('OverproductionWaste[excess_]', lambda excess: excess),
        ('InventoryWaste[excess_stock_]', lambda stock: stock),
        ('DefectWaste[rework_cost_]', lambda cost: cost),
        
        # ===== MAINTENANCE (20 functions) =====
        ('MTBF[operating_time_, failures_]', lambda t, f: t/f),  # Alias
        ('MeanTimeBetweenFailures[operating_time_, failures_]', lambda t, f: t/f),
        ('MTTR[repair_time_, repairs_]', lambda t, r: t/r),  # Alias
        ('MeanTimeToRepair[repair_time_, repairs_]', lambda t, r: t/r),
        ('MTTF[total_time_, items_]', lambda t, n: t/n),  # Alias
        ('MeanTimeToFailure[total_time_, items_]', lambda t, n: t/n),
        ('Availability[MTBF_, MTTR_]', lambda mtbf, mttr: mtbf/(mtbf + mttr)*100),
        ('ReliabilityFunction[failure_rate_, t_]', lambda lam, t: exp(-lam*t)),
        ('FailureRate[failures_, operating_time_]', lambda f, t: f/t),
        ('HazardFunction[h_]', lambda h: h),
        ('WeibullDistribution[t_, beta_, eta_]', lambda t, beta, eta: (beta/eta)*(t/eta)**(beta - 1)*exp(-(t/eta)**beta)),
        ('BathtubCurve[infant_, random_, wearout_]', lambda inf, rand, wear: inf + rand + wear),
        ('PreventiveMaintenanceCost[PM_cost_, frequency_]', lambda cost, freq: cost*freq),
        ('CorrectiveMaintenanceCost[failure_cost_, failures_]', lambda cost, fail: cost*fail),
        ('TotalMaintenanceCost[PM_, CM_]', lambda pm, cm: pm + cm),
        ('OptimalPMInterval[Cf_, Cp_, MTBF_]', lambda Cf, Cp, mtbf: mtbf*sqrt(Cp/Cf)),
        ('SparePartsInventory[failure_rate_, lead_time_]', lambda fr, lt: fr*lt),
        ('ConditionMonitoring[vibration_, temperature_, oil_]', lambda vib, temp, oil: (vib, temp, oil)),
        ('PredictiveMaintenance[trend_]', lambda trend: trend),
        ('MaintenanceBacklog[pending_work_]', lambda pw: pw),
        ('PlanningSchedulingEfficiency[planned_, total_]', lambda plan, total: plan/total*100),
        ('WorkOrderCompletion[completed_, total_]', lambda comp, total: comp/total*100),
        ('MaintenanceCostAsPercentRevenue[maint_cost_, revenue_]', lambda mc, rev: mc/rev*100),
        
        # ===== ERGONOMICS & SAFETY (20 functions) =====
        ('IncidentRate[incidents_, hours_worked_]', lambda inc, hrs: inc/hrs*200000),
        ('LostTimeInjuryRate[LTI_, hours_]', lambda lti, hrs: lti/hrs*1000000),
        ('TotalRecordableRate[TRC_, hours_]', lambda trc, hrs: trc/hrs*200000),
        ('SeverityRate[days_lost_, hours_]', lambda days, hrs: days/hrs*200000),
        ('FrequencyRate[accidents_, hours_]', lambda acc, hrs: acc/hrs*1000000),
        ('LiftingLimit[weight_, vertical_, horizontal_, asymmetry_]', lambda w, v, h, a: 51*(25/h)*(1 - 0.0075*abs(v - 30))*(1 - 0.0032*a)),
        ('NISOHRevisedLiftingEquation[LC_, HM_, VM_, DM_, AM_, FM_, CM_]', lambda LC, HM, VM, DM, AM, FM, CM: LC*HM*VM*DM*AM*FM*CM),
        ('LiftingIndex[load_, RWL_]', lambda load, rwl: load/rwl),
        ('MetabolicRate[activity_level_]', lambda act: act),
        ('EnergyExpenditure[met_, time_]', lambda met, t: met*t),
        ('RestAllowance[work_load_, duration_]', lambda wl, dur: wl*dur/100),
        ('NoiseExposureLevel[dB_, hours_]', lambda db, hrs: db + 10*log(hrs/8, 10)),
        ('TimeWeightedAverage[concentrations_, times_]', lambda C, t: sum(Ci*ti for Ci, ti in zip(C, t))/sum(t)),
        ('PermissibleExposureLimit[PEL_]', lambda pel: pel),
        ('ThresholdLimitValue[TLV_]', lambda tlv: tlv),
        ('HeatStressIndex[temp_, humidity_]', lambda T, H: 0.5*T + 0.5*H),
        ('WetBulbGlobeTemperature[WBGT_]', lambda wbgt: wbgt),
        ('IlluminationLevel[lux_]', lambda lux: lux),
        ('GlareRating[luminance_]', lambda L: L),
        ('AnthromorphicData[percentile_]', lambda p: p),
        
        # ===== FACILITY LAYOUT (20 functions) =====
        ('MaterialHandlingCost[distance_, trips_, cost_per_trip_]', lambda d, trips, cpt: d*trips*cpt),
        ('DistanceMatrix[locations_]', lambda locs: sp.Symbol('D')),
        ('FlowMatrix[from_to_]', lambda ft: sp.Symbol('F')),
        ('TotalMovementCost[flow_, distance_]', lambda F, D: sum(sum(Fij*Dij for Fij, Dij in zip(Fi, Di)) for Fi, Di in zip(F, D))),
        ('ClosenessRating[rating_]', lambda r: r),
        ('ActivityRelationshipChart[departments_]', lambda depts: sp.Symbol('ARC')),
        ('SpaceRequirement[machines_, area_per_machine_]', lambda m, a: m*a),
        ('FromToChart[trips_]', lambda trips: sp.Symbol('FT')),
        ('SystematicLayoutPlanning[SLP_]', lambda slp: slp),
        ('BlockLayout[blocks_]', lambda b: b),
        ('DetailedLayout[equipment_positions_]', lambda eq: eq),
        ('AisleWidth[traffic_, clearance_]', lambda traf, clear: traf + clear),
        ('ColumnSpacing[span_]', lambda span: span),
        ('CeilingHeight[equipment_, clearance_]', lambda eq, clear: eq + clear),
        ('LoadingDockCapacity[trucks_per_hour_]', lambda tph: tph),
        ('StorageCapacity[volume_]', lambda V: V),
        ('WarehouseUtilization[used_, total_]', lambda used, total: used/total*100),
        ('PickingEfficiency[orders_, time_]', lambda orders, t: orders/t),
        ('CrossDocking[incoming_, outgoing_]', lambda inc, out: inc == out),
        ('SlottingOptimization[ABC_classification_]', lambda abc: abc),
    ]

def register():
    """Register all extended36 rules"""
    from .extended_helper import convert_rules
    return convert_rules(get_rules())



"""Tests for the MikoshiLang parser — 50+ tests covering all syntax."""

import pytest
from mikoshilang.parser import parse, parse_file, tokenize, ParseError
from mikoshilang.expr import Expr, Symbol


class TestTokenizer:
    def test_integer(self):
        toks = tokenize("42")
        assert toks[0].type == "INTEGER" and toks[0].value == 42

    def test_float(self):
        toks = tokenize("3.14")
        assert toks[0].type == "FLOAT" and toks[0].value == 3.14

    def test_string(self):
        toks = tokenize('"hello"')
        assert toks[0].type == "STRING" and toks[0].value == "hello"

    def test_comment_stripped(self):
        toks = tokenize("(* comment *) 42")
        assert len(toks) == 1 and toks[0].value == 42

    def test_operators(self):
        toks = tokenize("+ - * / ^")
        types = [t.type for t in toks]
        assert types == ["PLUS", "MINUS", "TIMES", "DIVIDE", "POWER"]

    def test_brackets(self):
        toks = tokenize("[ ] { } ( )")
        types = [t.type for t in toks]
        assert types == ["LSQUARE", "RSQUARE", "LBRACE", "RBRACE", "LPAREN", "RPAREN"]


class TestArithmetic:
    def test_integer_literal(self):
        assert parse("42") == 42

    def test_float_literal(self):
        assert parse("3.14") == 3.14

    def test_negative_integer(self):
        assert parse("-5") == -5

    def test_addition(self):
        assert parse("2 + 3") == Expr("Plus", 2, 3)

    def test_subtraction(self):
        assert parse("5 - 3") == Expr("Plus", 5, Expr("Times", -1, 3))

    def test_multiplication(self):
        assert parse("2 * 3") == Expr("Times", 2, 3)

    def test_division(self):
        assert parse("6 / 2") == Expr("Times", 6, Expr("Power", 2, -1))

    def test_power(self):
        assert parse("x^2") == Expr("Power", Symbol("x"), 2)

    def test_precedence_mul_add(self):
        result = parse("2 + 3 * x")
        assert result == Expr("Plus", 2, Expr("Times", 3, Symbol("x")))

    def test_precedence_power_mul(self):
        result = parse("2 * x^3")
        assert result == Expr("Times", 2, Expr("Power", Symbol("x"), 3))

    def test_parentheses(self):
        result = parse("(2 + 3) * x")
        assert result == Expr("Times", Expr("Plus", 2, 3), Symbol("x"))

    def test_nested_parens(self):
        result = parse("((x))")
        assert result == Symbol("x")

    def test_complex_expression(self):
        result = parse("x^2 - 4")
        assert result == Expr("Plus", Expr("Power", Symbol("x"), 2), Expr("Times", -1, 4))


class TestImplicitMultiplication:
    def test_number_symbol(self):
        result = parse("2x")
        assert result == Expr("Times", 2, Symbol("x"))

    def test_number_function(self):
        result = parse("2Sin[x]")
        assert result == Expr("Times", 2, Expr("Sin", Symbol("x")))

    def test_paren_paren(self):
        result = parse("(x + 1)(x - 1)")
        lhs = Expr("Plus", Symbol("x"), 1)
        rhs = Expr("Plus", Symbol("x"), Expr("Times", -1, 1))
        assert result == Expr("Times", lhs, rhs)

    def test_number_paren(self):
        result = parse("3(x + 1)")
        assert result == Expr("Times", 3, Expr("Plus", Symbol("x"), 1))


class TestFunctionCalls:
    def test_sin(self):
        assert parse("Sin[x]") == Expr("Sin", Symbol("x"))

    def test_nested_function(self):
        assert parse("Sin[Cos[x]]") == Expr("Sin", Expr("Cos", Symbol("x")))

    def test_multiple_args(self):
        assert parse("Diff[x^2, x]") == Expr("Diff", Expr("Power", Symbol("x"), 2), Symbol("x"))

    def test_integrate(self):
        result = parse("Integrate[x^2, x]")
        assert result == Expr("Integrate", Expr("Power", Symbol("x"), 2), Symbol("x"))

    def test_solve(self):
        result = parse("Solve[x^2 - 4 == 0, x]")
        inner = Expr("Equal", Expr("Plus", Expr("Power", Symbol("x"), 2), Expr("Times", -1, 4)), 0)
        assert result == Expr("Solve", inner, Symbol("x"))

    def test_simplify(self):
        result = parse("Simplify[(x^2 - 1)/(x - 1)]")
        assert result.head == "Simplify"

    def test_factor(self):
        result = parse("Factor[x^2 - 4]")
        assert result.head == "Factor"

    def test_expand(self):
        result = parse("Expand[(x + 1)^3]")
        assert result.head == "Expand"

    def test_limit(self):
        result = parse("Limit[Sin[x]/x, x -> 0]")
        assert result.head == "Limit"

    def test_series(self):
        result = parse("Series[Exp[x], {x, 0, 5}]")
        assert result.head == "Series"
        assert isinstance(result.args[1], Expr) and result.args[1].head == "List"


class TestLists:
    def test_simple_list(self):
        result = parse("{1, 2, 3}")
        assert result == Expr("List", 1, 2, 3)

    def test_empty_list(self):
        result = parse("{}")
        assert result == Expr("List")

    def test_nested_list(self):
        result = parse("{{1, 2}, {3, 4}}")
        assert result == Expr("List", Expr("List", 1, 2), Expr("List", 3, 4))

    def test_list_with_exprs(self):
        result = parse("{x, x^2, x^3}")
        assert result.head == "List" and len(result.args) == 3

    def test_range(self):
        result = parse("Range[10]")
        assert result == Expr("Range", 10)

    def test_table(self):
        result = parse("Table[i^2, {i, 1, 10}]")
        assert result.head == "Table"

    def test_map(self):
        result = parse("Map[Sin, {1, 2, 3}]")
        assert result.head == "Map"

    def test_select(self):
        result = parse("Select[{1, -2, 3}, Positive]")
        assert result.head == "Select"


class TestMatrix:
    def test_matrix(self):
        result = parse("{{1, 2}, {3, 4}}")
        assert result == Expr("List", Expr("List", 1, 2), Expr("List", 3, 4))

    def test_det(self):
        result = parse("Det[{{1, 2}, {3, 4}}]")
        assert result.head == "Det"

    def test_inverse(self):
        result = parse("Inverse[{{1, 2}, {3, 4}}]")
        assert result.head == "Inverse"


class TestAssignmentAndRules:
    def test_set(self):
        result = parse("x = 5")
        assert result == Expr("Set", Symbol("x"), 5)

    def test_set_delayed(self):
        result = parse("f[x_] := x^2")
        assert result.head == "SetDelayed"

    def test_rule(self):
        result = parse("x -> 3")
        assert result == Expr("Rule", Symbol("x"), 3)

    def test_replace_all(self):
        result = parse("x /. x -> 3")
        assert result.head == "ReplaceAll"


class TestPatterns:
    def test_blank(self):
        result = parse("_")
        assert result == Expr("Blank")

    def test_named_blank(self):
        result = parse("x_")
        assert result == Expr("Pattern", Symbol("x"), Expr("Blank"))

    def test_typed_blank(self):
        result = parse("x_Integer")
        assert result == Expr("Pattern", Symbol("x"), Expr("Blank", Symbol("Integer")))

    def test_matchq(self):
        result = parse("MatchQ[Sin[x], Sin[_]]")
        assert result.head == "MatchQ"


class TestComparison:
    def test_equal(self):
        assert parse("x == y") == Expr("Equal", Symbol("x"), Symbol("y"))

    def test_unequal(self):
        assert parse("x != y") == Expr("Unequal", Symbol("x"), Symbol("y"))

    def test_greater(self):
        assert parse("x > 0") == Expr("Greater", Symbol("x"), 0)

    def test_less(self):
        assert parse("x < 0") == Expr("Less", Symbol("x"), 0)

    def test_geq(self):
        assert parse("x >= 0") == Expr("GreaterEqual", Symbol("x"), 0)

    def test_leq(self):
        assert parse("x <= 0") == Expr("LessEqual", Symbol("x"), 0)


class TestLogic:
    def test_and(self):
        assert parse("x && y") == Expr("And", Symbol("x"), Symbol("y"))

    def test_or(self):
        assert parse("x || y") == Expr("Or", Symbol("x"), Symbol("y"))

    def test_not(self):
        assert parse("!x") == Expr("Not", Symbol("x"))


class TestConstants:
    def test_pi(self):
        assert parse("Pi") == Symbol("Pi")

    def test_e(self):
        assert parse("E") == Symbol("E")

    def test_i(self):
        assert parse("I") == Symbol("I")

    def test_infinity(self):
        assert parse("Infinity") == Symbol("Infinity")

    def test_true(self):
        assert parse("True") is True

    def test_false(self):
        assert parse("False") is False


class TestEdgeCases:
    def test_empty_input(self):
        assert parse("") is None

    def test_comment_only(self):
        assert parse("(* just a comment *)") is None

    def test_string_literal(self):
        assert parse('"hello"') == "hello"

    def test_scientific_notation(self):
        result = parse("1e5")
        assert result == 1e5

    def test_multiple_operations(self):
        result = parse("1 + 2 + 3")
        assert result == Expr("Plus", Expr("Plus", 1, 2), 3)



"""Tests for numerical evaluation."""

import pytest
import math
from mikoshilang.expr import Expr, Symbol
from mikoshilang.numerical import N, numerical, Pi, E, I, Infinity, GoldenRatio


class TestConstants:
    def test_pi(self):
        assert abs(N(Pi) - math.pi) < 1e-10

    def test_e(self):
        assert abs(N(E) - math.e) < 1e-10

    def test_i(self):
        assert N(I) == complex(0, 1)

    def test_infinity(self):
        assert N(Infinity) == float("inf")

    def test_golden_ratio(self):
        phi = (1 + math.sqrt(5)) / 2
        assert abs(N(GoldenRatio) - phi) < 1e-10

    def test_pi_symbol(self):
        assert Pi.name == "Pi"

    def test_e_symbol(self):
        assert E.name == "E"


class TestNumerical:
    def test_int(self):
        assert N(42) == 42.0

    def test_float(self):
        assert N(3.14) == 3.14

    def test_symbol_unknown(self):
        x = Symbol("x")
        assert N(x) == x

    def test_expr_numeric(self):
        e = Expr("Plus", 1, 2)
        r = N(e)
        assert abs(r - 3.0) < 1e-10

    def test_sin_pi(self):
        e = Expr("Sin", Symbol("Pi"))
        r = N(e)
        assert abs(r) < 1e-10

    def test_cos_zero(self):
        e = Expr("Cos", 0)
        r = N(e)
        assert abs(r - 1.0) < 1e-10

    def test_exp_zero(self):
        e = Expr("Exp", 0)
        r = N(e)
        assert abs(r - 1.0) < 1e-10

    def test_log_e(self):
        e = Expr("Log", Symbol("E"))
        r = N(e)
        assert abs(r - 1.0) < 1e-10

    def test_power(self):
        e = Expr("Power", 2, 10)
        r = N(e)
        assert abs(r - 1024.0) < 1e-10

    def test_precision(self):
        r = N(Pi, 30)
        assert abs(r - math.pi) < 1e-14

    def test_numerical_alias(self):
        assert numerical(42) == N(42)

    def test_complex_expr(self):
        # sin(pi/4) ≈ sqrt(2)/2
        e = Expr("Sin", Expr("Times", Symbol("Pi"), Expr("Power", 4, -1)))
        r = N(e)
        assert abs(r - math.sqrt(2) / 2) < 1e-10



"""Tests for linear algebra operations."""

import pytest
from mikoshilang.expr import Expr
from mikoshilang.linalg import Matrix, Dot, Det, Inverse, Eigenvalues, Eigenvectors, Transpose


class TestMatrix:
    def test_create(self):
        m = Matrix([1, 2], [3, 4])
        assert m.head == "Matrix"
        assert len(m.args) == 2

    def test_repr(self):
        m = Matrix([1, 0], [0, 1])
        r = repr(m)
        assert "Matrix" in r

    def test_identity(self):
        m = Matrix([1, 0], [0, 1])
        assert m.args[0].args == (1, 0)
        assert m.args[1].args == (0, 1)


class TestDot:
    def test_identity(self):
        m = Matrix([1, 0], [0, 1])
        v = Matrix([3, 4], [5, 6])
        r = Dot(m, v)
        assert r.args[0].args == (3, 4)
        assert r.args[1].args == (5, 6)

    def test_multiplication(self):
        a = Matrix([1, 2], [3, 4])
        b = Matrix([5, 6], [7, 8])
        r = Dot(a, b)
        # [[1*5+2*7, 1*6+2*8], [3*5+4*7, 3*6+4*8]] = [[19,22],[43,50]]
        assert r.args[0].args == (19, 22)
        assert r.args[1].args == (43, 50)


class TestDet:
    def test_2x2(self):
        m = Matrix([1, 2], [3, 4])
        assert Det(m) == -2

    def test_identity(self):
        m = Matrix([1, 0], [0, 1])
        assert Det(m) == 1

    def test_singular(self):
        m = Matrix([1, 2], [2, 4])
        assert Det(m) == 0

    def test_3x3(self):
        m = Matrix([1, 2, 3], [4, 5, 6], [7, 8, 0])
        d = Det(m)
        assert d == 27


class TestInverse:
    def test_identity(self):
        m = Matrix([1, 0], [0, 1])
        r = Inverse(m)
        assert r.args[0].args == (1, 0)
        assert r.args[1].args == (0, 1)

    def test_2x2(self):
        m = Matrix([4, 7], [2, 6])
        r = Inverse(m)
        # Inverse of [[4,7],[2,6]] = [[0.6,-0.7],[-0.2,0.4]]
        assert abs(r.args[0].args[0] - 0.6) < 1e-10
        assert abs(r.args[0].args[1] - (-0.7)) < 1e-10


class TestEigenvalues:
    def test_diagonal(self):
        m = Matrix([2, 0], [0, 3])
        r = Eigenvalues(m)
        assert r.head == "List"
        assert set(r.args) == {2, 3}

    def test_identity(self):
        m = Matrix([1, 0], [0, 1])
        r = Eigenvalues(m)
        assert r.args == (1, 1)


class TestEigenvectors:
    def test_basic(self):
        m = Matrix([2, 0], [0, 3])
        r = Eigenvectors(m)
        assert r.head == "List"
        assert len(r.args) == 2


class TestTranspose:
    def test_2x2(self):
        m = Matrix([1, 2], [3, 4])
        r = Transpose(m)
        assert r.args[0].args == (1, 3)
        assert r.args[1].args == (2, 4)

    def test_identity(self):
        m = Matrix([1, 0], [0, 1])
        r = Transpose(m)
        assert r.args[0].args == (1, 0)

    def test_rectangular(self):
        m = Matrix([1, 2, 3], [4, 5, 6])
        r = Transpose(m)
        assert len(r.args) == 3
        assert r.args[0].args == (1, 4)



"""LaTeX rendering for MikoshiLang expressions."""

from __future__ import annotations
from .expr import Expr, Symbol


_GREEK = {
    "Alpha": r"\alpha", "Beta": r"\beta", "Gamma": r"\gamma", "Delta": r"\delta",
    "Epsilon": r"\epsilon", "Zeta": r"\zeta", "Eta": r"\eta", "Theta": r"\theta",
    "Iota": r"\iota", "Kappa": r"\kappa", "Lambda": r"\lambda", "Mu": r"\mu",
    "Nu": r"\nu", "Xi": r"\xi", "Pi": r"\pi", "Rho": r"\rho",
    "Sigma": r"\sigma", "Tau": r"\tau", "Upsilon": r"\upsilon", "Phi": r"\phi",
    "Chi": r"\chi", "Psi": r"\psi", "Omega": r"\omega",
    "alpha": r"\alpha", "beta": r"\beta", "gamma": r"\gamma", "delta": r"\delta",
    "epsilon": r"\epsilon", "zeta": r"\zeta", "eta": r"\eta", "theta": r"\theta",
    "pi": r"\pi", "sigma": r"\sigma", "tau": r"\tau", "phi": r"\phi",
    "omega": r"\omega",
}

_CONSTANTS = {
    "Pi": r"\pi", "E": r"e", "I": r"i", "Infinity": r"\infty",
    "GoldenRatio": r"\varphi",
}

_TRIG = {
    "Sin": r"\sin", "Cos": r"\cos", "Tan": r"\tan",
    "ArcSin": r"\arcsin", "ArcCos": r"\arccos", "ArcTan": r"\arctan",
    "Sinh": r"\sinh", "Cosh": r"\cosh", "Tanh": r"\tanh",
}


def to_latex(expr, parent_prec: int = 0) -> str:
    """Convert a MikoshiLang expression to a LaTeX string."""
    if isinstance(expr, bool):
        return r"\text{True}" if expr else r"\text{False}"
    if isinstance(expr, int):
        return str(expr)
    if isinstance(expr, float):
        return f"{expr:g}"
    if isinstance(expr, str):
        return r'\text{"' + expr + r'"}'
    if isinstance(expr, Symbol):
        name = expr.name
        if name in _CONSTANTS:
            return _CONSTANTS[name]
        if name in _GREEK:
            return _GREEK[name]
        if len(name) == 1:
            return name
        return r"\text{" + name + "}"
    if not isinstance(expr, Expr):
        return str(expr)

    head = expr.head
    args = expr.args

    if head == "Plus":
        parts = []
        for i, a in enumerate(args):
            is_neg_times = (isinstance(a, Expr) and a.head == "Times" and len(a.args) >= 1
                           and isinstance(a.args[0], (int, float)) and a.args[0] < 0)
            is_neg_atom = isinstance(a, (int, float)) and a < 0
            if i == 0:
                parts.append(to_latex(a, 1))
            else:
                if is_neg_atom:
                    parts.append(f" - {abs(a)}")
                elif is_neg_times:
                    coeff = a.args[0]
                    rest_args = a.args[1:]
                    if coeff == -1 and len(rest_args) == 1:
                        parts.append(f" - {to_latex(rest_args[0], 1)}")
                    else:
                        abs_coeff = abs(coeff)
                        if abs_coeff == 1 and len(rest_args) == 1:
                            pos_expr = rest_args[0]
                        else:
                            pos_expr = Expr("Times", abs_coeff, *rest_args)
                        parts.append(f" - {to_latex(pos_expr, 1)}")
                else:
                    parts.append(f" + {to_latex(a, 1)}")
        result = "".join(parts)
        if parent_prec > 1:
            result = r"\left(" + result + r"\right)"
        return result

    if head == "Times":
        # Check for negative coefficient
        if len(args) == 2 and args[0] == -1:
            inner = to_latex(args[1], 2)
            result = f"-{inner}"
            if parent_prec > 1:
                result = r"\left(" + result + r"\right)"
            return result
        # Check for fraction: a * b^-1
        num_parts = []
        den_parts = []
        for a in args:
            if isinstance(a, Expr) and a.head == "Power" and isinstance(a.args[1], int) and a.args[1] == -1:
                den_parts.append(a.args[0])
            elif isinstance(a, Expr) and a.head == "Power" and isinstance(a.args[1], (int, float)) and a.args[1] < 0:
                den_parts.append(Expr("Power", a.args[0], -a.args[1]))
            else:
                num_parts.append(a)
        if den_parts:
            num = _latex_product(num_parts) if num_parts else "1"
            den = _latex_product(den_parts)
            return r"\frac{" + num + "}{" + den + "}"
        return _latex_product(args)

    if head == "Power":
        base = args[0]
        exp = args[1]
        if isinstance(exp, Expr) and exp.head == "Times" and len(exp.args) == 2 and exp.args[0] == 1 and isinstance(exp.args[1], Expr) and exp.args[1].head == "Power" and exp.args[1].args[1] == -1:
            # x^(1/n) -> \sqrt[n]{x}
            n = exp.args[1].args[0]
            if n == 2:
                return r"\sqrt{" + to_latex(base) + "}"
            return r"\sqrt[" + to_latex(n) + "]{" + to_latex(base) + "}"
        if exp == -1:
            return r"\frac{1}{" + to_latex(base) + "}"
        if isinstance(exp, (int, float)) and exp == 0.5:
            return r"\sqrt{" + to_latex(base) + "}"
        base_s = to_latex(base, 4)
        exp_s = to_latex(exp)
        if _needs_braces(exp):
            return f"{base_s}^{{{exp_s}}}"
        return f"{base_s}^{{{exp_s}}}"

    if head in _TRIG:
        fn = _TRIG[head]
        arg = to_latex(args[0]) if args else ""
        return f"{fn}\\left({arg}\\right)"

    if head == "Exp":
        return r"e^{" + to_latex(args[0]) + "}"

    if head == "Log":
        if len(args) == 1:
            return r"\ln\left(" + to_latex(args[0]) + r"\right)"
        return r"\log_{" + to_latex(args[0]) + r"}\left(" + to_latex(args[1]) + r"\right)"

    if head == "Sqrt":
        return r"\sqrt{" + to_latex(args[0]) + "}"

    if head == "Abs":
        return r"\left|" + to_latex(args[0]) + r"\right|"

    if head == "Factorial":
        return to_latex(args[0], 4) + "!"

    if head == "List":
        inner = ", ".join(to_latex(a) for a in args)
        return r"\{" + inner + r"\}"

    if head == "Matrix":
        rows = []
        for row in args:
            if isinstance(row, Expr) and row.head == "List":
                rows.append(" & ".join(to_latex(x) for x in row.args))
            else:
                rows.append(to_latex(row))
        inner = r" \\ ".join(rows)
        return r"\begin{pmatrix} " + inner + r" \end{pmatrix}"

    if head == "Equal":
        return to_latex(args[0]) + " = " + to_latex(args[1])

    if head == "Unequal":
        return to_latex(args[0]) + r" \neq " + to_latex(args[1])

    if head == "Less":
        return to_latex(args[0]) + " < " + to_latex(args[1])

    if head == "Greater":
        return to_latex(args[0]) + " > " + to_latex(args[1])

    if head == "LessEqual":
        return to_latex(args[0]) + r" \leq " + to_latex(args[1])

    if head == "GreaterEqual":
        return to_latex(args[0]) + r" \geq " + to_latex(args[1])

    if head == "And":
        parts = [to_latex(a) for a in args]
        return r" \land ".join(parts)

    if head == "Or":
        parts = [to_latex(a) for a in args]
        return r" \lor ".join(parts)

    if head == "Not":
        return r"\lnot " + to_latex(args[0], 4)

    if head == "Integrate":
        if len(args) >= 2:
            return r"\int " + to_latex(args[0]) + r"\, d" + to_latex(args[1])
        return r"\int " + to_latex(args[0])

    if head == "Diff":
        if len(args) >= 2:
            return r"\frac{d}{d" + to_latex(args[1]) + "}" + to_latex(args[0])
        return r"\frac{d}{dx}" + to_latex(args[0])

    if head == "Limit":
        if len(args) >= 3:
            return r"\lim_{" + to_latex(args[1]) + r" \to " + to_latex(args[2]) + "} " + to_latex(args[0])
        return r"\lim " + to_latex(args[0])

    if head == "Sum":
        if len(args) >= 4:
            return r"\sum_{" + to_latex(args[1]) + "=" + to_latex(args[2]) + "}^{" + to_latex(args[3]) + "} " + to_latex(args[0])
        return r"\sum " + to_latex(args[0])

    if head == "Rule":
        return to_latex(args[0]) + r" \to " + to_latex(args[1])

    # Generic function
    inner = ", ".join(to_latex(a) for a in args)
    return r"\text{" + head + r"}\left(" + inner + r"\right)"


def _latex_product(parts) -> str:
    """Format a product of parts for LaTeX."""
    if not parts:
        return "1"
    strs = []
    for p in parts:
        s = to_latex(p, 2)
        strs.append(s)
    return " ".join(strs)


def _needs_braces(expr) -> bool:
    """Check if expression needs braces in superscript."""
    if isinstance(expr, (int, float)):
        s = str(expr)
        return len(s) > 1 or expr < 0
    return True



"""Tests for pattern matching."""

import pytest
from mikoshilang.expr import Expr, Symbol
from mikoshilang.pattern import Blank, BlankSequence, BlankNullSequence, Pattern, Condition, match, parse_pattern


class TestBlank:
    def test_match_int(self):
        assert match(Blank(), 42) == {}

    def test_match_symbol(self):
        x = Symbol("x")
        assert match(Blank(), x) == {}

    def test_match_expr(self):
        e = Expr("Plus", 1, 2)
        assert match(Blank(), e) == {}

    def test_named_blank(self):
        r = match(Blank("x"), 42)
        assert r == {"x": 42}

    def test_named_blank_consistent(self):
        # Same name must bind same value
        pat = Expr("Plus", Blank("x"), Blank("x"))
        assert match(pat, Expr("Plus", 3, 3)) == {"x": 3}
        assert match(pat, Expr("Plus", 3, 4)) is None

    def test_head_constraint_integer(self):
        assert match(Blank(head="Integer"), 42) == {}
        assert match(Blank(head="Integer"), 3.14) is None
        assert match(Blank(head="Integer"), Symbol("x")) is None

    def test_head_constraint_real(self):
        assert match(Blank(head="Real"), 3.14) == {}
        assert match(Blank(head="Real"), 42) is None

    def test_head_constraint_string(self):
        assert match(Blank(head="String"), "hello") == {}
        assert match(Blank(head="String"), 42) is None

    def test_head_constraint_symbol(self):
        assert match(Blank(head="Symbol"), Symbol("x")) == {}
        assert match(Blank(head="Symbol"), 42) is None

    def test_head_constraint_expr(self):
        assert match(Blank(head="Plus"), Expr("Plus", 1, 2)) == {}
        assert match(Blank(head="Plus"), Expr("Times", 1, 2)) is None

    def test_named_with_head(self):
        r = match(Blank("x", "Integer"), 42)
        assert r == {"x": 42}
        assert match(Blank("x", "Integer"), 3.14) is None


class TestPatternMatch:
    def test_exact_int(self):
        assert match(42, 42) == {}
        assert match(42, 43) is None

    def test_exact_symbol(self):
        x = Symbol("x")
        assert match(x, x) == {}
        assert match(x, Symbol("y")) is None

    def test_expr_match(self):
        pat = Expr("Plus", Blank("a"), Blank("b"))
        r = match(pat, Expr("Plus", 1, 2))
        assert r == {"a": 1, "b": 2}

    def test_expr_head_mismatch(self):
        pat = Expr("Plus", Blank("a"), Blank("b"))
        assert match(pat, Expr("Times", 1, 2)) is None

    def test_nested_pattern(self):
        pat = Expr("Plus", Expr("Times", Blank("a"), Blank("b")), Blank("c"))
        r = match(pat, Expr("Plus", Expr("Times", 2, 3), 4))
        assert r == {"a": 2, "b": 3, "c": 4}

    def test_pattern_object(self):
        pat = Pattern("Plus", Blank("a"), Blank("b"))
        r = match(pat, Expr("Plus", 1, 2))
        assert r == {"a": 1, "b": 2}


class TestBlankSequence:
    def test_match_one(self):
        pat = Expr("f", BlankSequence("xs"))
        r = match(pat, Expr("f", 1))
        assert r is not None
        assert r["xs"] == (1,)

    def test_match_many(self):
        pat = Expr("f", BlankSequence("xs"))
        r = match(pat, Expr("f", 1, 2, 3))
        assert r is not None
        assert r["xs"] == (1, 2, 3)

    def test_no_match_empty(self):
        pat = Expr("f", BlankSequence("xs"))
        assert match(pat, Expr("f")) is None


class TestBlankNullSequence:
    def test_match_zero(self):
        pat = Expr("f", BlankNullSequence("xs"))
        r = match(pat, Expr("f"))
        assert r is not None
        assert r["xs"] == ()

    def test_match_many(self):
        pat = Expr("f", BlankNullSequence("xs"))
        r = match(pat, Expr("f", 1, 2))
        assert r is not None
        assert r["xs"] == (1, 2)

    def test_with_fixed_arg(self):
        pat = Expr("f", Blank("a"), BlankNullSequence("rest"))
        r = match(pat, Expr("f", 1, 2, 3))
        assert r == {"a": 1, "rest": (2, 3)}

    def test_with_fixed_arg_min(self):
        pat = Expr("f", Blank("a"), BlankNullSequence("rest"))
        r = match(pat, Expr("f", 1))
        assert r == {"a": 1, "rest": ()}


class TestCondition:
    def test_condition_pass(self):
        pat = Condition(Blank("x"), lambda b: isinstance(b["x"], int) and b["x"] > 0)
        assert match(pat, 5) == {"x": 5}

    def test_condition_fail(self):
        pat = Condition(Blank("x"), lambda b: isinstance(b["x"], int) and b["x"] > 0)
        assert match(pat, -1) is None

    def test_condition_fail_type(self):
        pat = Condition(Blank("x"), lambda b: isinstance(b["x"], int) and b["x"] > 0)
        assert match(pat, 3.5) is None


class TestParsePattern:
    def test_blank(self):
        p = parse_pattern("_")
        assert isinstance(p, Blank)
        assert p.name is None

    def test_named_blank(self):
        p = parse_pattern("x_")
        assert isinstance(p, Blank)
        assert p.name == "x"

    def test_typed_blank(self):
        p = parse_pattern("x_Integer")
        assert isinstance(p, Blank)
        assert p.name == "x"
        assert p.head == "Integer"

    def test_blank_sequence(self):
        p = parse_pattern("__")
        assert isinstance(p, BlankSequence)

    def test_named_blank_sequence(self):
        p = parse_pattern("xs__")
        assert isinstance(p, BlankSequence)
        assert p.name == "xs"

    def test_null_sequence(self):
        p = parse_pattern("___")
        assert isinstance(p, BlankNullSequence)

    def test_named_null_sequence(self):
        p = parse_pattern("xs___")
        assert isinstance(p, BlankNullSequence)
        assert p.name == "xs"

    def test_symbol(self):
        p = parse_pattern("x")
        assert isinstance(p, Symbol)
        assert p.name == "x"



"""Data operations — lists, tables, maps, and statistics."""

from __future__ import annotations
from typing import Any, Callable, Optional
import statistics as _stats
from .expr import Expr, Symbol


def List(*args) -> Expr:
    """Create an ordered list."""
    return Expr("List", *args)


def Table(expr_fn: Callable, spec) -> Expr:
    """Generate a list. spec = (min, max) or (min, max, step)."""
    if isinstance(spec, (list, tuple)):
        if len(spec) == 2:
            start, end = spec
            step = 1
        elif len(spec) == 3:
            start, end, step = spec
        else:
            raise ValueError("Table spec must be (min, max) or (min, max, step)")
    else:
        start, end, step = 1, spec, 1
    results = []
    i = start
    while i <= end:
        results.append(expr_fn(i))
        i += step
    return Expr("List", *results)


def Map(fn: Callable, lst: Expr) -> Expr:
    """Apply fn to each element of a List."""
    if isinstance(lst, Expr) and lst.head == "List":
        return Expr("List", *(fn(a) for a in lst.args))
    raise TypeError("Map expects a List expression")


def Select(lst: Expr, cond: Callable) -> Expr:
    """Filter elements of a List by condition."""
    if isinstance(lst, Expr) and lst.head == "List":
        return Expr("List", *(a for a in lst.args if cond(a)))
    raise TypeError("Select expects a List expression")


def Sort(lst: Expr, key=None) -> Expr:
    """Sort a List."""
    if isinstance(lst, Expr) and lst.head == "List":
        return Expr("List", *sorted(lst.args, key=key))
    raise TypeError("Sort expects a List expression")


def Reverse(lst: Expr) -> Expr:
    """Reverse a List."""
    if isinstance(lst, Expr) and lst.head == "List":
        return Expr("List", *reversed(lst.args))
    raise TypeError("Reverse expects a List expression")


def Take(lst: Expr, n: int) -> Expr:
    """Take first n elements."""
    if isinstance(lst, Expr) and lst.head == "List":
        return Expr("List", *lst.args[:n])
    raise TypeError("Take expects a List expression")


def Drop(lst: Expr, n: int) -> Expr:
    """Drop first n elements."""
    if isinstance(lst, Expr) and lst.head == "List":
        return Expr("List", *lst.args[n:])
    raise TypeError("Drop expects a List expression")


def Part(lst: Expr, *indices: int) -> Any:
    """Access element(s) by index (1-based like Wolfram)."""
    if isinstance(lst, Expr) and lst.head == "List":
        result = lst
        for idx in indices:
            if isinstance(result, Expr) and result.head == "List":
                result = result.args[idx - 1] if idx > 0 else result.args[idx]
            else:
                raise IndexError(f"Cannot index into {result}")
        return result
    raise TypeError("Part expects a List expression")


def Range(*args) -> Expr:
    """Range(n), Range(a, b), or Range(a, b, step)."""
    if len(args) == 1:
        return Expr("List", *builtins_range(1, args[0] + 1))
    elif len(args) == 2:
        return Expr("List", *builtins_range(args[0], args[1] + 1))
    elif len(args) == 3:
        return Expr("List", *builtins_range(args[0], args[1] + 1, args[2]))
    raise ValueError("Range takes 1-3 arguments")


def builtins_range(start, stop, step=1):
    """Python range as list of ints."""
    return list(range(start, stop, step))


def Total(lst: Expr) -> Any:
    """Sum of elements."""
    if isinstance(lst, Expr) and lst.head == "List":
        return sum(lst.args)
    raise TypeError("Total expects a List expression")


def Mean(lst: Expr) -> float:
    """Arithmetic mean."""
    if isinstance(lst, Expr) and lst.head == "List":
        return _stats.mean(lst.args)
    raise TypeError("Mean expects a List expression")


def Median(lst: Expr) -> float:
    """Median value."""
    if isinstance(lst, Expr) and lst.head == "List":
        return _stats.median(lst.args)
    raise TypeError("Median expects a List expression")


def StandardDeviation(lst: Expr) -> float:
    """Standard deviation (sample)."""
    if isinstance(lst, Expr) and lst.head == "List":
        return _stats.stdev(lst.args)
    raise TypeError("StandardDeviation expects a List expression")



"""
MikoshiLang Extended Functions - Set 6
Advanced calculus, geometry, topology, and tensor operations
"""
import sympy as sp
from sympy import symbols, Function, diff, integrate, Matrix, simplify, expand, sqrt, sin, cos, tan, pi, E, I

def get_rules():
    """Extended set 6: Advanced mathematical domains (250 functions)"""
    x, y, z, t, u, v, w = symbols('x y z t u v w')
    f, g, h = symbols('f g h', cls=Function)
    
    return [
        # ===== VECTOR CALCULUS (40 functions) =====
        ('Grad[f_, {vars__}]', lambda expr, *vars: Matrix([diff(expr, v) for v in vars])),
        ('Gradient[f_, {vars__}]', lambda expr, *vars: Matrix([diff(expr, v) for v in vars])),
        ('Div[{f__}, {vars__}]', lambda *args: sum(diff(args[i], args[-len(args)//2+i]) for i in range(len(args)//2))),
        ('Divergence[{f__}, {vars__}]', lambda *args: sum(diff(args[i], args[-len(args)//2+i]) for i in range(len(args)//2))),
        ('Curl[{fx_, fy_, fz_}, {x_, y_, z_}]', lambda fx, fy, fz, x, y, z: Matrix([
            diff(fz, y) - diff(fy, z),
            diff(fx, z) - diff(fz, x),
            diff(fy, x) - diff(fx, y)
        ])),
        ('Laplacian[f_, {vars__}]', lambda expr, *vars: sum(diff(expr, v, 2) for v in vars)),
        ('VectorPotential[{fx_, fy_, fz_}, {x_, y_, z_}]', lambda fx, fy, fz, x, y, z: Matrix([
            integrate(fz, y) - integrate(fy, z),
            integrate(fx, z) - integrate(fz, x),
            integrate(fy, x) - integrate(fx, y)
        ])),
        ('ScalarPotential[{fx_, fy_, fz_}, {x_, y_, z_}]', lambda fx, fy, fz, x, y, z: integrate(fx, x) + integrate(fy, y) + integrate(fz, z)),
        ('Hessian[f_, {vars__}]', lambda expr, *vars: Matrix([[diff(expr, vi, vj) for vj in vars] for vi in vars])),
        ('Jacobian[{f__}, {vars__}]', lambda *args: Matrix([[diff(fi, vj) for vj in args[len(args)//2:]] for fi in args[:len(args)//2]])),
        ('DirectionalDerivative[f_, {vars__}, {dir__}]', lambda expr, vars, dirs: sum(diff(expr, v) * d for v, d in zip(vars, dirs))),
        ('NormalVector[f_, {x_, y_}]', lambda f, x, y: Matrix([diff(f, x), diff(f, y), -1]) / sqrt(diff(f, x)**2 + diff(f, y)**2 + 1)),
        ('TangentVector[{x_, y_, z_}, t_]', lambda x, y, z, t: Matrix([diff(x, t), diff(y, t), diff(z, t)])),
        ('Curvature[{x_, y_}, t_]', lambda x, y, t: abs(diff(x, t) * diff(y, t, 2) - diff(y, t) * diff(x, t, 2)) / (diff(x, t)**2 + diff(y, t)**2)**(sp.Rational(3,2))),
        ('Torsion[{x_, y_, z_}, t_]', lambda x, y, z, t: sp.det(Matrix([
            [diff(x, t), diff(y, t), diff(z, t)],
            [diff(x, t, 2), diff(y, t, 2), diff(z, t, 2)],
            [diff(x, t, 3), diff(y, t, 3), diff(z, t, 3)]
        ])) / (diff(x, t)**2 + diff(y, t)**2 + diff(z, t)**2)**3),
        ('ArcLength[{x_, y_}, {t_, a_, b_}]', lambda x, y, t, a, b: integrate(sqrt(diff(x, t)**2 + diff(y, t)**2), (t, a, b))),
        ('SurfaceArea[f_, {x_, a_, b_}, {y_, c_, d_}]', lambda f, x, a, b, y, c, d: integrate(
            integrate(sqrt(1 + diff(f, x)**2 + diff(f, y)**2), (x, a, b)), (y, c, d))),
        ('Volume[f_, {x_, a_, b_}, {y_, c_, d_}, {z_, e_, f_}]', lambda *args: integrate(integrate(integrate(args[0], (args[1], args[2], args[3])), (args[4], args[5], args[6])), (args[7], args[8], args[9]))),
        ('FluxIntegral[{fx_, fy_, fz_}, surf_]', lambda fx, fy, fz, surf: integrate(integrate(fx + fy + fz, surf), surf)),
        ('WorkIntegral[{fx_, fy_}, path_]', lambda fx, fy, path: integrate(fx + fy, path)),
        
        # ===== DIFFERENTIAL GEOMETRY (35 functions) =====
        ('Christoffel[g_, {vars__}]', lambda g, *vars: sp.tensorcontraction(
            sp.tensorproduct(g.inv(), sum(diff(g, v) for v in vars)), (0, 1))),
        ('RiemannTensor[g_, {vars__}]', lambda g, *vars: Matrix([[diff(diff(g[i,j], vars[k]), vars[l]) - 
            diff(diff(g[i,j], vars[l]), vars[k]) for l in range(len(vars))] for k in range(len(vars)) for j in range(len(vars)) for i in range(len(vars))])),
        ('RicciTensor[R_]', lambda R: sp.tensorcontraction(R, (0, 2))),
        ('RicciScalar[Ric_]', lambda Ric: sp.trace(Ric)),
        ('EinsteinTensor[Ric_, R_, g_]', lambda Ric, R, g: Ric - R * g / 2),
        ('WeylTensor[R_, Ric_, R_scalar_, g_]', lambda R, Ric, Rs, g: R - (Ric - Rs * g / 2)),
        ('MetricTensor[{coords__}]', lambda *coords: Matrix([[sp.KroneckerDelta(i, j) for j in range(len(coords))] for i in range(len(coords))])),
        ('FirstFundamentalForm[{x_, y_, z_}, {u_, v_}]', lambda x, y, z, u, v: Matrix([
            [diff(x, u)**2 + diff(y, u)**2 + diff(z, u)**2, diff(x, u)*diff(x, v) + diff(y, u)*diff(y, v) + diff(z, u)*diff(z, v)],
            [diff(x, u)*diff(x, v) + diff(y, u)*diff(y, v) + diff(z, u)*diff(z, v), diff(x, v)**2 + diff(y, v)**2 + diff(z, v)**2]
        ])),
        ('SecondFundamentalForm[{x_, y_, z_}, {u_, v_}]', lambda x, y, z, u, v: Matrix([
            [diff(x, u, 2), diff(x, u, v)],
            [diff(x, u, v), diff(x, v, 2)]
        ])),
        ('GaussianCurvature[I_, II_]', lambda I, II: sp.det(II) / sp.det(I)),
        ('MeanCurvature[I_, II_]', lambda I, II: sp.trace(I.inv() * II) / 2),
        ('PrincipalCurvatures[I_, II_]', lambda I, II: (I.inv() * II).eigenvals().keys()),
        ('ShapeOperator[I_, II_]', lambda I, II: I.inv() * II),
        ('CovariantDerivative[T_, g_, {vars__}]', lambda T, g, *vars: diff(T, vars[0]) + 
            sum(sp.symbols('Gamma^i_jk') * T for _ in vars)),
        ('ParallelTransport[v_, path_]', lambda v, path: v),  # Simplified
        
        # ===== TENSOR OPERATIONS (40 functions) =====
        ('TensorProduct[A_, B_]', lambda A, B: sp.tensorproduct(A, B)),
        ('TensorContraction[T_, {i_, j_}]', lambda T, i, j: sp.tensorcontraction(T, (i, j))),
        ('OuterProduct[v_, w_]', lambda v, w: sp.tensorproduct(v, w)),
        ('InnerProduct[v_, w_]', lambda v, w: v.dot(w) if hasattr(v, 'dot') else sum(vi*wi for vi, wi in zip(v, w))),
        ('WedgeProduct[v_, w_]', lambda v, w: v[0]*w[1] - v[1]*w[0] if len(v) == 2 else Matrix([
            v[1]*w[2] - v[2]*w[1], v[2]*w[0] - v[0]*w[2], v[0]*w[1] - v[1]*w[0]
        ])),
        ('HodgeDual[form_]', lambda form: form),  # Simplified
        ('ExteriorDerivative[form_, {vars__}]', lambda form, *vars: sum(diff(form, v) for v in vars)),
        ('LieBracket[X_, Y_]', lambda X, Y: X * Y - Y * X),
        ('SymmetricPart[T_]', lambda T: (T + T.T) / 2 if hasattr(T, 'T') else T),
        ('AntisymmetricPart[T_]', lambda T: (T - T.T) / 2 if hasattr(T, 'T') else T),
        ('Symmetrize[T_]', lambda T: (T + T.T) / 2 if hasattr(T, 'T') else T),
        ('Antisymmetrize[T_]', lambda T: (T - T.T) / 2 if hasattr(T, 'T') else T),
        
        # ===== GEOMETRY (45 functions) =====
        ('Distance2D[{x1_, y1_}, {x2_, y2_}]', lambda x1, y1, x2, y2: sqrt((x2-x1)**2 + (y2-y1)**2)),
        ('Distance3D[{x1_, y1_, z1_}, {x2_, y2_, z2_}]', lambda x1, y1, z1, x2, y2, z2: sqrt((x2-x1)**2 + (y2-y1)**2 + (z2-z1)**2)),
        ('Midpoint[{x1_, y1_}, {x2_, y2_}]', lambda x1, y1, x2, y2: ((x1+x2)/2, (y1+y2)/2)),
        ('Slope[{x1_, y1_}, {x2_, y2_}]', lambda x1, y1, x2, y2: (y2-y1)/(x2-x1) if x2 != x1 else sp.oo),
        ('PointLineDistance[{x0_, y0_}, {a_, b_, c_}]', lambda x0, y0, a, b, c: abs(a*x0 + b*y0 + c) / sqrt(a**2 + b**2)),
        ('AngleBetweenLines[m1_, m2_]', lambda m1, m2: sp.atan(abs((m2-m1)/(1+m1*m2)))),
        ('CircleArea[r_]', lambda r: pi * r**2),
        ('CircleCircumference[r_]', lambda r: 2 * pi * r),
        ('SphereVolume[r_]', lambda r: 4 * pi * r**3 / 3),
        ('SphereSurfaceArea[r_]', lambda r: 4 * pi * r**2),
        ('CylinderVolume[r_, h_]', lambda r, h: pi * r**2 * h),
        ('ConeVolume[r_, h_]', lambda r, h: pi * r**2 * h / 3),
        ('TriangleArea[{x1_, y1_}, {x2_, y2_}, {x3_, y3_}]', lambda x1, y1, x2, y2, x3, y3: 
            abs(x1*(y2-y3) + x2*(y3-y1) + x3*(y1-y2)) / 2),
        ('PolygonArea[{pts__}]', lambda *pts: sum(pts[i][0]*(pts[(i+1)%len(pts)][1] - pts[(i-1)%len(pts)][1]) 
            for i in range(len(pts))) / 2),
        ('Centroid[{pts__}]', lambda *pts: (sum(p[0] for p in pts)/len(pts), sum(p[1] for p in pts)/len(pts))),
        ('ConvexHull[{pts__}]', lambda *pts: pts),  # Simplified
        ('VoronoiDiagram[{pts__}]', lambda *pts: pts),  # Simplified
        ('DelaunayTriangulation[{pts__}]', lambda *pts: pts),  # Simplified
        
        # ===== TOPOLOGY (30 functions) =====
        ('EulerCharacteristic[V_, E_, F_]', lambda V, E, F: V - E + F),
        ('BettiNumber[n_, complex_]', lambda n, c: 0),  # Placeholder
        ('FundamentalGroup[space_]', lambda s: 'Z'),  # Placeholder
        ('HomologyGroup[n_, space_]', lambda n, s: 'Z'),  # Placeholder
        ('CohomologyGroup[n_, space_]', lambda n, s: 'Z'),  # Placeholder
        ('ChainComplex[simplices_]', lambda s: s),
        ('BoundaryOperator[simplex_]', lambda s: s),
        ('Coboundary[cochain_]', lambda c: c),
        ('SimplexOrientation[simplex_]', lambda s: 1),
        ('SimplicialComplex[simplices_]', lambda s: s),
        ('NerveOfCover[cover_]', lambda c: c),
        
        # ===== FRACTALS & CHAOS (25 functions) =====
        ('MandelbrotSet[c_, maxiter_]', lambda c, n: sum(1 for _ in range(int(n)) if abs(c) < 2)),
        ('JuliaSet[z_, c_, maxiter_]', lambda z, c, n: sum(1 for _ in range(int(n)) if abs(z) < 2)),
        ('LogisticMap[r_, x_]', lambda r, x: r * x * (1 - x)),
        ('LyapunovExponent[f_, x0_, n_]', lambda f, x0, n: sum(sp.log(abs(diff(f, x))) for _ in range(int(n))) / n),
        ('BifurcationDiagram[f_, param_, range_]', lambda f, p, r: [(p, f) for p in r]),
        ('AttractorDimension[pts_]', lambda pts: len(pts)),
        ('PoincarePlot[traj_]', lambda t: t),
        ('RecurrencePlot[series_]', lambda s: Matrix([[1 if abs(s[i]-s[j]) < 0.1 else 0 for j in range(len(s))] for i in range(len(s))])),
        
        # ===== SPECIAL COORDINATES (35 functions) =====
        ('PolarToCartesian[r_, theta_]', lambda r, th: (r*cos(th), r*sin(th))),
        ('CartesianToPolar[x_, y_]', lambda x, y: (sqrt(x**2 + y**2), sp.atan2(y, x))),
        ('CylindricalToCartesian[r_, theta_, z_]', lambda r, th, z: (r*cos(th), r*sin(th), z)),
        ('CartesianToCylindrical[x_, y_, z_]', lambda x, y, z: (sqrt(x**2 + y**2), sp.atan2(y, x), z)),
        ('SphericalToCartesian[r_, theta_, phi_]', lambda r, th, ph: 
            (r*sin(ph)*cos(th), r*sin(ph)*sin(th), r*cos(ph))),
        ('CartesianToSpherical[x_, y_, z_]', lambda x, y, z: 
            (sqrt(x**2 + y**2 + z**2), sp.atan2(y, x), sp.acos(z/sqrt(x**2 + y**2 + z**2)))),
        ('ToroidalToCartesian[R_, r_, theta_, phi_]', lambda R, r, th, ph: 
            ((R + r*cos(ph))*cos(th), (R + r*cos(ph))*sin(th), r*sin(ph))),
        ('ParabolicToCartesian[u_, v_]', lambda u, v: ((u**2 - v**2)/2, u*v)),
        ('EllipticToCartesian[a_, mu_, nu_]', lambda a, mu, nu: 
            (a*sp.cosh(mu)*cos(nu), a*sp.sinh(mu)*sin(nu))),
        ('BipolarToCartesian[a_, u_, v_]', lambda a, u, v: 
            (a*sp.sinh(u)/(sp.cosh(u) - cos(v)), a*sin(v)/(sp.cosh(u) - cos(v)))),
        
        # ===== COMPUTATIONAL GEOMETRY (40 functions) =====
        ('PointInPolygon[{x_, y_}, polygon_]', lambda x, y, poly: True),  # Placeholder
        ('LineIntersection[{x1_, y1_}, {x2_, y2_}, {x3_, y3_}, {x4_, y4_}]', lambda x1, y1, x2, y2, x3, y3, x4, y4:
            ((x1*y2 - y1*x2)*(x3-x4) - (x1-x2)*(x3*y4 - y3*x4)) / ((x1-x2)*(y3-y4) - (y1-y2)*(x3-x4))),
        ('PolygonIntersection[poly1_, poly2_]', lambda p1, p2: []),  # Placeholder
        ('ClosestPair[{pts__}]', lambda *pts: (pts[0], pts[1]) if len(pts) > 1 else None),
        ('FarthestPair[{pts__}]', lambda *pts: (pts[0], pts[-1]) if len(pts) > 1 else None),
        ('RotatePoint[{x_, y_}, angle_]', lambda x, y, a: (x*cos(a) - y*sin(a), x*sin(a) + y*cos(a))),
        ('ScalePoint[{x_, y_}, factor_]', lambda x, y, f: (x*f, y*f)),
        ('TranslatePoint[{x_, y_}, {dx_, dy_}]', lambda x, y, dx, dy: (x+dx, y+dy)),
        ('ReflectPoint[{x_, y_}, axis_]', lambda x, y, ax: (-x, y) if ax == 'y' else (x, -y)),
        ('BoundingBox[{pts__}]', lambda *pts: (min(p[0] for p in pts), min(p[1] for p in pts), 
            max(p[0] for p in pts), max(p[1] for p in pts))),
        ('MinimumEnclosingCircle[{pts__}]', lambda *pts: (sum(p[0] for p in pts)/len(pts), 
            sum(p[1] for p in pts)/len(pts), max(sqrt((p[0]-sum(q[0] for q in pts)/len(pts))**2 + 
            (p[1]-sum(q[1] for q in pts)/len(pts))**2) for p in pts))),
    ]


def register():
    """Register all extended6 rules"""
    from .extended_helper import convert_rules
    return convert_rules(get_rules())



"""
MikoshiLang Knowledge Layer - Entity Graph + Text Retrieval

Architecture:
1. Entity layer (Wikidata) - structured facts, QIDs, properties
2. Text layer (Wikipedia) - summaries, context, citations
3. Deterministic tools - no LLM hallucination
4. Provenance tracking - sources + timestamps

License compliance:
- Wikidata: CC0 (public domain)
- Wikipedia: CC BY-SA 3.0 (attribution required)
"""

import requests
import json
import time
from datetime import datetime, timezone
from functools import lru_cache
from typing import Dict, List, Optional, Tuple, Any

# Wikidata API endpoints
WIKIDATA_API = "https://www.wikidata.org/w/api.php"
WIKIPEDIA_API = "https://en.wikipedia.org/w/api.php"

# User-Agent header (required by Wikimedia APIs)
HEADERS = {
    "User-Agent": "MikoshiLang/1.0 (https://github.com/DarrenEdwards111/MikoshiLang; mikoshi.ai) Python/3.x"
}

# Canonical property mappings (Mikoshi name → Wikidata PID)
CANONICAL_PROPERTIES = {
    "BirthDate": "P569",
    "DeathDate": "P570",
    "BirthPlace": "P19",
    "Occupation": "P106",
    "InstanceOf": "P31",
    "SubclassOf": "P279",
    "Country": "P17",
    "Population": "P1082",
    "Coordinate": "P625",
    "Mass": "P2067",
    "Height": "P2048",
    "Director": "P57",
    "Author": "P50",
    "Publisher": "P123",
    "PublicationDate": "P577",
    "ISBN": "P212",
    "DOI": "P356",
    "ChemicalFormula": "P274",
    "AtomicNumber": "P1086",
    "MolarMass": "P2067",
}

# Reverse mapping (PID → human-readable name)
PID_TO_NAME = {v: k for k, v in CANONICAL_PROPERTIES.items()}

# Cache for entity lookups (keyed by query + type)
@lru_cache(maxsize=1000)
def _entity_search_cached(query: str, entity_type: Optional[str], limit: int) -> str:
    """Cached Wikidata entity search"""
    # If entity_type provided, append it to the search query as a hint
    search_query = f"{query} {entity_type}" if entity_type else query
    
    params = {
        "action": "wbsearchentities",
        "format": "json",
        "language": "en",
        "search": search_query,
        "limit": limit,
    }
    
    try:
        response = requests.get(WIKIDATA_API, params=params, headers=HEADERS, timeout=10)
        response.raise_for_status()
        return response.text
    except Exception as e:
        return json.dumps({"error": str(e)})


def EntitySearch(query: str, entity_type: Optional[str] = None, limit: int = 5) -> List[Dict[str, Any]]:
    """
    Search for entities in Wikidata.
    
    Args:
        query: Search string (e.g., "Douglas Adams", "Moon")
        entity_type: Optional type hint ("Person", "Film", "Place", etc.)
        limit: Max results to return
    
    Returns:
        List of candidate entities with:
        - id: Wikidata QID (e.g., "Q42")
        - label: Primary name
        - description: Short description
        - url: Wikidata URL
        - retrieved: Timestamp
    
    Example:
        >>> EntitySearch["Douglas Adams", "Person"]
        [{"id": "Q42", "label": "Douglas Adams", 
          "description": "English writer and humorist", ...}]
    """
    response_text = _entity_search_cached(query, entity_type, limit)
    data = json.loads(response_text)
    
    if "error" in data:
        return [{"error": data["error"]}]
    
    results = []
    for item in data.get("search", []):
        results.append({
            "id": item.get("id"),
            "label": item.get("label"),
            "description": item.get("description", ""),
            "url": f"https://www.wikidata.org/wiki/{item.get('id')}",
            "retrieved": datetime.now(timezone.utc).isoformat() + "Z",
            "source": "Wikidata (CC0)",
        })
    
    return results


@lru_cache(maxsize=500)
def _entity_data_cached(qid: str) -> str:
    """Cached Wikidata entity data fetch"""
    params = {
        "action": "wbgetentities",
        "format": "json",
        "ids": qid,
        "languages": "en",
    }
    
    try:
        response = requests.get(WIKIDATA_API, params=params, headers=HEADERS, timeout=10)
        response.raise_for_status()
        return response.text
    except Exception as e:
        return json.dumps({"error": str(e)})


def EntityValue(entity_id: str, property_name: str) -> Dict[str, Any]:
    """
    Get a specific property value for an entity.
    
    Args:
        entity_id: Wikidata QID (e.g., "Q42") or entity dict from EntitySearch
        property_name: Canonical property name (e.g., "BirthDate", "Occupation")
                      or Wikidata PID (e.g., "P569")
    
    Returns:
        Dict with:
        - value: The property value(s)
        - property: Property name
        - pid: Wikidata property ID
        - source: Source URL
        - retrieved: Timestamp
    
    Example:
        >>> EntityValue["Q42", "BirthDate"]
        {"value": "1952-03-11", "property": "BirthDate", "pid": "P569", ...}
    """
    # Handle entity dict input
    if isinstance(entity_id, dict):
        entity_id = entity_id.get("id")
    
    # Resolve property name to PID
    if property_name.startswith("P") and property_name[1:].isdigit():
        pid = property_name
        prop_name = PID_TO_NAME.get(pid, pid)
    else:
        pid = CANONICAL_PROPERTIES.get(property_name, property_name)
        prop_name = property_name
    
    # Fetch entity data
    response_text = _entity_data_cached(entity_id)
    data = json.loads(response_text)
    
    if "error" in data:
        return {"error": data["error"]}
    
    entity_data = data.get("entities", {}).get(entity_id, {})
    claims = entity_data.get("claims", {}).get(pid, [])
    
    if not claims:
        return {
            "value": None,
            "property": prop_name,
            "pid": pid,
            "entity": entity_id,
            "error": f"Property {prop_name} ({pid}) not found",
        }
    
    # Extract values
    values = []
    for claim in claims:
        mainsnak = claim.get("mainsnak", {})
        datavalue = mainsnak.get("datavalue", {})
        
        if datavalue.get("type") == "time":
            # Time values (dates)
            time_val = datavalue.get("value", {}).get("time", "")
            # Strip leading + and precision info
            time_val = time_val.lstrip("+").split("T")[0]
            values.append(time_val)
        elif datavalue.get("type") == "wikibase-entityid":
            # Entity references (linked items)
            linked_id = datavalue.get("value", {}).get("id")
            values.append({"id": linked_id, "url": f"https://www.wikidata.org/wiki/{linked_id}"})
        elif datavalue.get("type") == "string":
            values.append(datavalue.get("value"))
        elif datavalue.get("type") == "quantity":
            amount = datavalue.get("value", {}).get("amount")
            unit = datavalue.get("value", {}).get("unit", "")
            values.append({"amount": amount, "unit": unit})
        elif datavalue.get("type") == "globecoordinate":
            coord = datavalue.get("value", {})
            values.append({
                "latitude": coord.get("latitude"),
                "longitude": coord.get("longitude"),
            })
        else:
            values.append(str(datavalue.get("value")))
    
    # Return single value if only one, else list
    final_value = values[0] if len(values) == 1 else values
    
    return {
        "value": final_value,
        "property": prop_name,
        "pid": pid,
        "entity": entity_id,
        "source": f"https://www.wikidata.org/wiki/{entity_id}",
        "retrieved": datetime.now(timezone.utc).isoformat() + "Z",
        "license": "CC0 (Wikidata)",
    }


def EntityProperties(entity_id: str) -> Dict[str, Any]:
    """
    Get all available properties for an entity.
    
    Args:
        entity_id: Wikidata QID or entity dict
    
    Returns:
        Dict mapping property names to values
    
    Example:
        >>> EntityProperties["Q42"]
        {"BirthDate": "1952-03-11", "Occupation": [...], ...}
    """
    if isinstance(entity_id, dict):
        entity_id = entity_id.get("id")
    
    response_text = _entity_data_cached(entity_id)
    data = json.loads(response_text)
    
    if "error" in data:
        return {"error": data["error"]}
    
    entity_data = data.get("entities", {}).get(entity_id, {})
    claims = entity_data.get("claims", {})
    
    properties = {}
    for pid, claim_list in claims.items():
        # Use canonical name if available, else PID
        prop_name = PID_TO_NAME.get(pid, pid)
        result = EntityValue(entity_id, pid)
        if "error" not in result:
            properties[prop_name] = result["value"]
    
    return {
        "entity": entity_id,
        "properties": properties,
        "source": f"https://www.wikidata.org/wiki/{entity_id}",
        "retrieved": datetime.now(timezone.utc).isoformat() + "Z",
    }


@lru_cache(maxsize=500)
def _wikipedia_summary_cached(title: str) -> str:
    """Cached Wikipedia summary fetch"""
    params = {
        "action": "query",
        "format": "json",
        "titles": title,
        "prop": "extracts",
        "exintro": True,
        "explaintext": True,
    }
    
    try:
        response = requests.get(WIKIPEDIA_API, params=params, headers=HEADERS, timeout=10)
        response.raise_for_status()
        return response.text
    except Exception as e:
        return json.dumps({"error": str(e)})


def WikipediaSummary(title: str, sentences: int = 3) -> Dict[str, Any]:
    """
    Get Wikipedia article summary (lead section).
    
    Args:
        title: Wikipedia article title
        sentences: Number of sentences to return (approximate)
    
    Returns:
        Dict with:
        - summary: Text summary
        - title: Article title
        - url: Wikipedia URL
        - retrieved: Timestamp
        - license: CC BY-SA 3.0
    
    Example:
        >>> WikipediaSummary["Douglas Adams"]
        {"summary": "Douglas Noel Adams was an English author...", ...}
    """
    response_text = _wikipedia_summary_cached(title)
    data = json.loads(response_text)
    
    if "error" in data:
        return {"error": data["error"]}
    
    pages = data.get("query", {}).get("pages", {})
    page = next(iter(pages.values()), {})
    
    if "missing" in page:
        return {"error": f"Wikipedia article '{title}' not found"}
    
    extract = page.get("extract", "")
    
    # Truncate to requested sentences (rough)
    sentences_list = extract.split(". ")[:sentences]
    summary = ". ".join(sentences_list)
    if not summary.endswith("."):
        summary += "."
    
    return {
        "summary": summary,
        "title": page.get("title"),
        "url": f"https://en.wikipedia.org/wiki/{page.get('title', '').replace(' ', '_')}",
        "retrieved": datetime.now(timezone.utc).isoformat() + "Z",
        "license": "CC BY-SA 3.0 (Wikipedia)",
        "attribution": "Source: Wikipedia contributors",
    }


def Disambiguate(query: str, context: Optional[str] = None) -> List[Dict[str, Any]]:
    """
    Find disambiguation candidates for ambiguous queries.
    
    Args:
        query: Ambiguous term (e.g., "Moon")
        context: Optional context hint (e.g., "film", "astronomy")
    
    Returns:
        List of candidates with descriptions for user to choose
    
    Example:
        >>> Disambiguate["Moon", "film"]
        [{"id": "Q219562", "label": "Moon", "description": "2009 film by Duncan Jones"}, ...]
    """
    # If context provided, use it in search
    search_query = f"{query} {context}" if context else query
    
    # Search Wikidata
    entities = EntitySearch(search_query, limit=10)
    
    # Also check for Wikipedia disambiguation page
    wiki_params = {
        "action": "query",
        "format": "json",
        "titles": f"{query} (disambiguation)",
        "prop": "revisions",
        "rvprop": "content",
        "rvslots": "main",
    }
    
    try:
        wiki_response = requests.get(WIKIPEDIA_API, params=wiki_params, timeout=10)
        wiki_data = wiki_response.json()
        wiki_pages = wiki_data.get("query", {}).get("pages", {})
        
        # If disambiguation page exists, note it
        has_disambig = any("disambiguation" in str(p.get("title", "")).lower() 
                          for p in wiki_pages.values())
    except:
        has_disambig = False
    
    return {
        "query": query,
        "context": context,
        "candidates": entities,
        "has_disambiguation_page": has_disambig,
        "note": "Multiple entities found - please refine query or select by ID",
    }


# Convenience function for natural language queries
def Entity(name_or_type: str, name: Optional[str] = None) -> Dict[str, Any]:
    """
    Convenience wrapper for entity lookup.
    
    Usage:
        Entity["Douglas Adams"]           # Search by name
        Entity["Person", "Douglas Adams"] # Search with type hint
    
    Returns the top entity match or disambiguation prompt.
    """
    if name is None:
        # Single argument: just a name
        results = EntitySearch(name_or_type, limit=5)
    else:
        # Two arguments: type, name
        results = EntitySearch(name, entity_type=name_or_type, limit=5)
    
    if not results:
        return {"error": "No entities found"}
    
    # Always return the first result if we have results
    return results[0]


# Export all knowledge functions
__all__ = [
    "EntitySearch",
    "EntityValue",
    "EntityProperties",
    "WikipediaSummary",
    "Disambiguate",
    "Entity",
    "CANONICAL_PROPERTIES",
]



"""Signal processing module for MikoshiLang.

Wraps numpy.fft and scipy.signal for numerical operations,
SymPy for symbolic Fourier transforms.
"""

from __future__ import annotations
from typing import Any, List as ListType, Optional
import numpy as np
from .expr import Expr, Symbol


# ── Symbolic Fourier transforms (via SymPy) ──

def FourierTransform(expr, t_var, w_var) -> Any:
    """Symbolic Fourier transform using SymPy."""
    import sympy as sp
    from .math_ops import to_sympy, from_sympy
    sexpr = to_sympy(expr)
    st = to_sympy(t_var) if isinstance(t_var, Symbol) else sp.Symbol(str(t_var))
    sw = to_sympy(w_var) if isinstance(w_var, Symbol) else sp.Symbol(str(w_var))
    result = sp.fourier_transform(sexpr, st, sw)
    return from_sympy(result)


def InverseFourierTransform(expr, w_var, t_var) -> Any:
    """Symbolic inverse Fourier transform using SymPy."""
    import sympy as sp
    from .math_ops import to_sympy, from_sympy
    sexpr = to_sympy(expr)
    sw = to_sympy(w_var) if isinstance(w_var, Symbol) else sp.Symbol(str(w_var))
    st = to_sympy(t_var) if isinstance(t_var, Symbol) else sp.Symbol(str(t_var))
    result = sp.inverse_fourier_transform(sexpr, sw, st)
    return from_sympy(result)


# ── Discrete transforms ──

def _list_to_array(lst: Expr) -> np.ndarray:
    """Convert a List Expr to numpy array."""
    if isinstance(lst, Expr) and lst.head == "List":
        return np.array([complex(x) if isinstance(x, complex) else float(x) for x in lst.args])
    raise TypeError("Expected a List expression")


def _array_to_list(arr: np.ndarray) -> Expr:
    """Convert numpy array to List Expr."""
    result = []
    for x in arr:
        if np.isreal(x):
            v = float(np.real(x))
            if v == int(v):
                result.append(int(v))
            else:
                result.append(v)
        else:
            result.append(complex(x))
    return Expr("List", *result)


def DFT(data: Expr) -> Expr:
    """Discrete Fourier Transform."""
    arr = _list_to_array(data)
    result = np.fft.fft(arr)
    return _array_to_list(result)


def IDFT(data: Expr) -> Expr:
    """Inverse Discrete Fourier Transform."""
    arr = _list_to_array(data)
    result = np.fft.ifft(arr)
    return _array_to_list(result)


# ── Filters ──

def LowPassFilter(data: Expr, cutoff: float, sample_rate: float = 1.0, order: int = 5) -> Expr:
    """Apply a Butterworth low-pass filter."""
    from scipy.signal import butter, filtfilt
    arr = _list_to_array(data).real
    nyq = 0.5 * sample_rate
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff, btype='low')
    filtered = filtfilt(b, a, arr)
    return _array_to_list(filtered)


def HighPassFilter(data: Expr, cutoff: float, sample_rate: float = 1.0, order: int = 5) -> Expr:
    """Apply a Butterworth high-pass filter."""
    from scipy.signal import butter, filtfilt
    arr = _list_to_array(data).real
    nyq = 0.5 * sample_rate
    normal_cutoff = cutoff / nyq
    b, a = butter(order, normal_cutoff, btype='high')
    filtered = filtfilt(b, a, arr)
    return _array_to_list(filtered)


def BandPassFilter(data: Expr, low: float, high: float, sample_rate: float = 1.0, order: int = 5) -> Expr:
    """Apply a Butterworth band-pass filter."""
    from scipy.signal import butter, filtfilt
    arr = _list_to_array(data).real
    nyq = 0.5 * sample_rate
    low_n = low / nyq
    high_n = high / nyq
    b, a = butter(order, [low_n, high_n], btype='band')
    filtered = filtfilt(b, a, arr)
    return _array_to_list(filtered)


# ── Convolution ──

def Convolve(signal1: Expr, signal2: Expr) -> Expr:
    """Convolve two signals."""
    arr1 = _list_to_array(signal1).real
    arr2 = _list_to_array(signal2).real
    result = np.convolve(arr1, arr2)
    return _array_to_list(result)


# ── Window functions ──

def HammingWindow(n: int) -> Expr:
    """Generate a Hamming window of length n."""
    return _array_to_list(np.hamming(n))


def HanningWindow(n: int) -> Expr:
    """Generate a Hanning window of length n."""
    return _array_to_list(np.hanning(n))


def BlackmanWindow(n: int) -> Expr:
    """Generate a Blackman window of length n."""
    return _array_to_list(np.blackman(n))


# ── Spectrogram ──

def Spectrogram(data: Expr, sample_rate: float = 1.0) -> Any:
    """Compute spectrogram. Returns matplotlib figure if available, else raw data."""
    from scipy.signal import spectrogram as scipy_spectrogram
    arr = _list_to_array(data).real
    f, t, Sxx = scipy_spectrogram(arr, fs=sample_rate)
    try:
        import matplotlib
        matplotlib.use("Agg")
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots()
        ax.pcolormesh(t, f, 10 * np.log10(Sxx + 1e-10))
        ax.set_ylabel("Frequency")
        ax.set_xlabel("Time")
        ax.set_title("Spectrogram")
        fig.tight_layout()
        return fig
    except ImportError:
        return Expr("List",
                     Expr("Rule", "frequencies", _array_to_list(f)),
                     Expr("Rule", "times", _array_to_list(t)))



"""
MikoshiLang Extended Functions - Set 24
Environmental science, ecology, and geoscience
"""
import sympy as sp
from sympy import symbols, sqrt, exp, log, pi

def get_rules():
    """Extended set 24: Environmental and earth sciences (150 functions)"""
    t, C, x = symbols('t C x')
    
    return [
        # ===== ECOLOGY (40 functions) =====
        ('SpeciesRichness[S_]', lambda S: S),
        ('ShannonDiversityIndex[p_i_]', lambda p: -sum(pi*log(pi) for pi in p if pi > 0)),
        ('SimpsonDiversityIndex[p_i_]', lambda p: 1 - sum(pi**2 for pi in p)),
        ('PielouEvennessIndex[H_, S_]', lambda H, S: H/log(S)),
        ('BergerParkerIndex[N_max_, N_total_]', lambda Nmax, Ntot: Nmax/Ntot),
        ('BrayCurtisDistance[site1_, site2_]', lambda s1, s2: sum(abs(s1i - s2i) for s1i, s2i in zip(s1, s2))/sum(s1i + s2i for s1i, s2i in zip(s1, s2))),
        ('JaccardIndex[set1_, set2_]', lambda s1, s2: len(set(s1) & set(s2))/len(set(s1) | set(s2))),
        ('SorensenIndex[set1_, set2_]', lambda s1, s2: 2*len(set(s1) & set(s2))/(len(set(s1)) + len(set(s2)))),
        ('MorisitaIndex[n1_, n2_, N1_, N2_]', lambda n1, n2, N1, N2: 2*sum(n1i*n2i for n1i, n2i in zip(n1, n2))/(sum(n1i**2 for n1i in n1)/N1**2 + sum(n2i**2 for n2i in n2)/N2**2)/(N1*N2)),
        ('RarefactionCurve[individuals_, species_]', lambda ind, spp: spp),
        ('SpeciesAreaCurve[area_, z_, c_]', lambda A, z, c: c*A**z),
        ('MacArthurWilsonModel[immigration_, extinction_]', lambda I, E: I == E),
        ('IslandBiogeography[distance_, area_]', lambda d, A: log(A)/d),
        ('NicheBreadth[p_i_]', lambda p: 1/sum(pi**2 for pi in p)),
        ('NicheOverlap[p_i_, q_i_]', lambda p, q: sum(pi*qi for pi, qi in zip(p, q))),
        ('ResourcePartitioning[niches_]', lambda n: len(n)),
        ('CarryingCapacity[resources_, consumption_]', lambda R, cons: R/cons),
        ('LogisticGrowthEcology[r_, K_, N_]', lambda r, K, N: r*N*(1 - N/K)),
        ('ExponentialGrowthEcology[r_, N_]', lambda r, N: r*N),
        ('PopulationDoublingTimeEcology[r_]', lambda r: log(2)/r),
        ('GenerationTime[age_, l_x_, m_x_]', lambda age, lx, mx: sum(agei*lxi*mxi for agei, lxi, mxi in zip(age, lx, mx))/sum(lxi*mxi for lxi, mxi in zip(lx, mx))),
        ('NetReproductiveRate[l_x_, m_x_]', lambda lx, mx: sum(lxi*mxi for lxi, mxi in zip(lx, mx))),
        ('IntrinsicRateOfIncrease[r_]', lambda r: r),
        ('LifeTable[age_, n_x_, l_x_, d_x_, q_x_]', lambda age, nx, lx, dx, qx: list(zip(age, nx, lx, dx, qx))),
        ('SurvivalCurve[l_x_]', lambda lx: lx),
        ('Mortality Rate[d_x_, n_x_]', lambda dx, nx: dx/nx),
        ('Fecundity[offspring_, females_]', lambda off, fem: off/fem),
        ('AgeDist ribution[n_x_]', lambda nx: [nxi/sum(nx) for nxi in nx]),
        ('StableAgeDistribution[leslie_matrix_]', lambda L: sp.Symbol('eigenvector')),
        ('LeslieMatrix[survival_, fecundity_]', lambda s, f: sp.Matrix([[f[0], f[1]], [s[0], 0]])),
        ('MetapopulationModel[patches_, migration_]', lambda p, m: p),
        ('SourceSinkDynamics[birth_, death_, immigration_, emigration_]', lambda b, d, i, e: b - d + i - e),
        ('RescueEffect[extinction_prob_, colonization_]', lambda pe, col: pe*(1 - col)),
        ('TrophicLevel[position_]', lambda pos: pos),
        ('FoodChainLength[levels_]', lambda lev: len(lev)),
        ('EnergyTransferEfficiency[E_out_, E_in_]', lambda Eout, Ein: Eout/Ein*100),
        ('BiomagnificationFactor[C_predator_, C_prey_]', lambda Cpred, Cprey: Cpred/Cprey),
        ('EcologicalEfficiency[production_n1_, production_n_]', lambda Pn1, Pn: Pn1/Pn*100),
        ('NetPrimaryProductivity[GPP_, R_]', lambda GPP, R: GPP - R),
        ('GrossPrimaryProductivity[photosynthesis_]', lambda P: P),
        
        # ===== BIOGEOCHEMICAL CYCLES (30 functions) =====
        ('CarbonCycle[photosynthesis_, respiration_, decomposition_]', lambda P, R, D: P - R - D),
        ('NitrogenFixation[rate_]', lambda r: r),
        ('Nitrification[NH4_, NO3_]', lambda NH4, NO3: NH4 - NO3),
        ('Denitrification[NO3_, N2_]', lambda NO3, N2: NO3 - N2),
        ('PhosphorusCycle[weathering_, uptake_, mineralization_]', lambda w, u, m: w - u + m),
        ('SulfurCycle[volcanic_, biological_]', lambda v, b: v + b),
        ('WaterCycle[precipitation_, evaporation_, runoff_]', lambda P, E, R: P - E - R),
        ('Evapotranspiration[evaporation_, transpiration_]', lambda E, T: E + T),
        ('Infiltration[precip_, runoff_, evap_]', lambda P, R, E: P - R - E),
        ('GroundwaterRecharge[infiltration_, ET_]', lambda I, ET: I - ET),
        ('StreamFlow[precipitation_, ET_, storage_change_]', lambda P, ET, dS: P - ET - dS),
        ('SedimentYield[erosion_, deposition_]', lambda eros, dep: eros - dep),
        ('WeatheringRate[temp_, precip_, rock_type_]', lambda T, P, rock: T*P*sp.Symbol('k_rock')),
        ('SoilFormation[weathering_, transport_]', lambda w, t: w - t),
        ('DecompositionRate[litter_, k_]', lambda litter, k: litter*k),
        ('CarbonSequestration[uptake_, release_]', lambda u, r: u - r),
        ('MethaneFlux[production_, oxidation_]', lambda prod, ox: prod - ox),
        ('N2OEmissions[nitrification_, denitrification_]', lambda nit, denit: nit + denit),
        ('DOCConcentration[organic_matter_]', lambda OM: OM*sp.Symbol('solubility')),
        ('DICConcentration[CO2_, HCO3_, CO3_]', lambda CO2, HCO3, CO3: CO2 + HCO3 + CO3),
        ('AlkalinityTotal[HCO3_, CO3_, OH_, H_]', lambda HCO3, CO3, OH, H: HCO3 + 2*CO3 + OH - H),
        ('pHFromCO2[CO2_, Kh_]', lambda CO2, Kh: -log(sqrt(Kh*CO2), 10)),
        ('BufferCapacity[alkalinity_]', lambda alk: alk),
        ('CarbonateSpeciation[pH_, DIC_]', lambda pH, DIC: (sp.Symbol('CO2'), sp.Symbol('HCO3'), sp.Symbol('CO3'))),
        ('CalciteSaturation[Ca_, CO3_, Ksp_]', lambda Ca, CO3, Ksp: Ca*CO3/Ksp),
        ('AragnoiteSaturation[Ca_, CO3_, Ksp_]', lambda Ca, CO3, Ksp: Ca*CO3/Ksp),
        ('OceanAcidification[pH_past_, pH_present_]', lambda pHp, pHpr: pHp - pHpr),
        ('CO2Solubility[T_, S_]', lambda T, S: exp(-60.2409 + 93.4517*(100/T) + 23.3585*log(T/100) + S*(0.023517 - 0.023656*(T/100) + 0.0047036*(T/100)**2))),
        ('GasExchange[k_, delta_C_]', lambda k, dC: k*dC),
        ('PistonVelocity[wind_, Sc_]', lambda u, Sc: 0.31*u**2*(Sc/660)**(-0.5)),
        
        # ===== POLLUTION & TOXICOLOGY (30 functions) =====
        ('AdvectionDispersion[C_, v_, D_]', lambda C, v, D: v*sp.diff(C, sp.Symbol('x')) - D*sp.diff(C, sp.Symbol('x'), 2)),
        ('DilutionFactor[C_source_, C_ambient_]', lambda Cs, Ca: Cs/Ca),
        ('MixingZone[discharge_, velocity_, depth_]', lambda Q, v, h: Q/(v*h)),
        ('PlumеDispersion[Q_, u_, sigma_y_, sigma_z_]', lambda Q, u, sigy, sigz: Q/(2*pi*u*sigy*sigz)),
        ('GaussianPlume[x_, y_, z_, H_, u_, sigma_y_, sigma_z_]', lambda x, y, z, H, u, sigy, sigz: exp(-y**2/(2*sigy**2))*(exp(-(z - H)**2/(2*sigz**2)) + exp(-(z + H)**2/(2*sigz**2)))/(2*pi*u*sigy*sigz)),
        ('AirQualityIndex[PM25_, PM10_, O3_, NO2_, SO2_, CO_]', lambda pm25, pm10, o3, no2, so2, co: max(pm25, pm10, o3, no2, so2, co)),
        ('WaterQualityIndex[DO_, BOD_, pH_, turbidity_]', lambda DO, BOD, pH, turb: (DO + 1/BOD + pH + 1/turb)/4),
        ('BiologicalOxygenDemand[C0_, k_, t_]', lambda C0, k, t: C0*(1 - exp(-k*t))),
        ('ChemicalOxygenDemand[organic_matter_]', lambda OM: OM*sp.Symbol('oxidation_factor')),
        ('DissolvedOxygen[saturation_, deficit_]', lambda sat, def_: sat - def_),
        ('OxygenSaturation[T_, S_]', lambda T, S: 14.6 - 0.41*T - 0.008*S),  # mg/L approximation
        ('OxygenDeficit[DO_sat_, DO_actual_]', lambda DOsat, DOact: DOsat - DOact),
        ('StreeterPhelps[La_, ka_, kr_, t_]', lambda La, ka, kr, t: La*ka/(kr - ka)*(exp(-ka*t) - exp(-kr*t))),
        ('CriticalDeficit[La_, ka_, kr_]', lambda La, ka, kr: La*ka/kr*(kr/ka)**(kr/(kr - ka))),
        ('Reaeration[k2_, deficit_]', lambda k2, D: k2*D),
        ('Deoxygenation[k1_, BOD_]', lambda k1, BOD: k1*BOD),
        ('Bioaccumulation[C_water_, BCF_]', lambda Cw, BCF: Cw*BCF),
        ('BioconcentrationFactor[C_organism_, C_water_]', lambda Co, Cw: Co/Cw),
        ('Bioaccumulation Factor[C_organism_, C_water_]', lambda Co, Cw: Co/Cw),
        ('BioamplificationFactor[C_predator_, C_prey_]', lambda Cpred, Cprey: Cpred/Cprey),
        ('PersistenceIndex[half_life_]', lambda t12: log(2)/t12),
        ('DegradationRate[C0_, C_, t_]', lambda C0, C, t: -log(C/C0)/t),
        ('HalfLifePollutant[k_]', lambda k: log(2)/k),
        ('ResidenceTime[volume_, flow_]', lambda V, Q: V/Q),
        ('RetentionTime[mass_, input_rate_]', lambda M, I: M/I),
        ('LoadingRate[mass_, area_, time_]', lambda M, A, t: M/(A*t)),
        ('RemovalEfficiency[C_in_, C_out_]', lambda Cin, Cout: (Cin - Cout)/Cin*100),
        ('TreatmentEfficiency[removed_, input_]', lambda rem, inp: rem/inp*100),
        ('SludgeProductionRate[BOD_removed_, yield_]', lambda BODrem, Y: BODrem*Y),
        ('HydraulicRetentionTime[V_, Q_]', lambda V, Q: V/Q),
        
        # ===== SOIL SCIENCE (25 functions) =====
        ('BulkDensitySoil[mass_dry_, volume_total_]', lambda md, Vt: md/Vt),
        ('ParticleDensity[mass_, volume_solid_]', lambda m, Vs: m/Vs),
        ('Porosity[bulk_density_, particle_density_]', lambda rhob, rhop: 1 - rhob/rhop),
        ('VoidRatioSoil[porosity_]', lambda n: n/(1 - n)),
        ('WaterContentGravimetric[mass_water_, mass_dry_]', lambda mw, md: mw/md),
        ('WaterContentVolumetric[volume_water_, volume_total_]', lambda Vw, Vt: Vw/Vt),
        ('SaturationDegree[theta_, porosity_]', lambda theta, n: theta/n),
        ('FieldCapacity[water_content_]', lambda theta: theta),  # At -33 kPa
        ('WiltingPoint[water_content_]', lambda theta: theta),  # At -1500 kPa
        ('AvailableWaterCapacity[FC_, WP_]', lambda FC, WP: FC - WP),
        ('SoilWaterPotential[psi_]', lambda psi: psi),
        ('MatricPotential[tension_]', lambda h: -h),
        ('HydraulicConductivity[K_]', lambda K: K),
        ('InfiltrationRate[i_]', lambda i: i),
        ('HortonEquation[fc_, f0_, k_, t_]', lambda fc, f0, k, t: fc + (f0 - fc)*exp(-k*t)),
        ('GreenAmptInfiltration[K_, psi_, delta_theta_, F_]', lambda K, psi, dtheta, F: K*(1 + psi*dtheta/F)),
        ('DarcysLaw[K_, dh_dl_]', lambda K, dhdl: -K*dhdl),
        ('SoilMoistureCharacteristic[theta_, psi_]', lambda theta, psi: (theta, psi)),
        ('VanGenuchtenModel[theta_s_, theta_r_, alpha_, n_, psi_]', lambda thetas, thetar, alpha, n, psi: thetar + (thetas - thetar)/(1 + (alpha*abs(psi))**n)**(1 - 1/n)),
        ('BrooksCoreyModel[theta_s_, theta_r_, psi_b_, lambda_, psi_]', lambda thetas, thetar, psib, lam, psi: thetar + (thetas - thetar)*(psi/psib)**(-lam)),
        ('SoilpH[H_concentration_]', lambda H: -log(H, 10)),
        ('CationExchangeCapacity[cmol_kg_]', lambda CEC: CEC),
        ('BaseSaturation[exchangeable_bases_, CEC_]', lambda bases, CEC: bases/CEC*100),
        ('SodiumAdsorptionRatio[Na_, Ca_, Mg_]', lambda Na, Ca, Mg: Na/sqrt((Ca + Mg)/2)),
        ('ElectricalConductivity[salinity_]', lambda EC: EC),  # dS/m
        
        # ===== GEOPHYSICS (25 functions) =====
        ('SeismicWaveVelocity[P_wave_, S_wave_]', lambda Vp, Vs: (Vp, Vs)),
        ('PWaveVelocity[K_, mu_, rho_]', lambda K, mu, rho: sqrt((K + 4*mu/3)/rho)),
        ('SWaveVelocity[mu_, rho_]', lambda mu, rho: sqrt(mu/rho)),
        ('PoissonRatioSeismic[Vp_, Vs_]', lambda Vp, Vs: (Vp**2 - 2*Vs**2)/(2*(Vp**2 - Vs**2))),
        ('RayleighWaveVelocity[Vs_]', lambda Vs: 0.92*Vs),
        ('LoveWaveVelocity[Vs_]', lambda Vs: Vs),
        ('EarthquakeMagnitude[amplitude_, distance_]', lambda A, d: log(A, 10) + 3*log(d, 10)),
        ('RichterScale[amplitude_]', lambda A: log(A, 10)),
        ('MomentMagnitude[M0_]', lambda M0: sp.Rational(2,3)*log(M0, 10) - 10.7),
        ('SeismicMoment[mu_, A_, D_]', lambda mu, A, D: mu*A*D),
        ('EnergyRelease[magnitude_]', lambda M: 10**(1.5*M + 4.8)),
        ('FocalDepth[d_]', lambda d: d),
        ('Epicentraldistance[delta_t_, v_]', lambda dt, v: dt*v),
        ('TravelTime[distance_, velocity_]', lambda d, v: d/v),
        ('SnellsLaw[v1_, v2_, i1_]', lambda v1, v2, i1: sp.asin(v2/v1*sin(i1))),
        ('ReflectionCoefficient[Z1_, Z2_]', lambda Z1, Z2: (Z2 - Z1)/(Z2 + Z1)),
        ('TransmissionCoefficient[Z1_, Z2_]', lambda Z1, Z2: 2*Z2/(Z2 + Z1)),
        ('AcousticImpedance[rho_, v_]', lambda rho, v: rho*v),
        ('GeothermalGradient[delta_T_, delta_z_]', lambda dT, dz: dT/dz),
        ('HeatFlowDensity[k_, dT_dz_]', lambda k, dTdz: -k*dTdz),
        ('ThermalConductivity[heat_flow_, gradient_]', lambda q, grad: q/grad),
        ('GravityAnomaly[g_observed_, g_theoretical_]', lambda gobs, gth: gobs - gth),
        ('BouguerAnomaly[g_obs_, g_normal_, g_free_air_, g_bouguer_]', lambda gobs, gn, gfa, gb: gobs - gn - gfa - gb),
        ('IsostasyAiry[h_, rho_crust_, rho_mantle_]', lambda h, rhoc, rhom: h*rhoc/(rhom - rhoc)),
        ('IsostasyPratt[h1_, h2_, rho1_, rho2_]', lambda h1, h2, rho1, rho2: h1*rho1 == h2*rho2),
    ]


def register():
    """Register all extended24 rules"""
    from .extended_helper import convert_rules
    return convert_rules(get_rules())



"""
MikoshiLang Extended Functions - Set 29
Information theory, cryptography, coding theory, and data compression
"""
import sympy as sp
from sympy import symbols, sqrt, exp, log, pi, factorial

def get_rules():
    """Extended set 29: Information theory and cryptography (150 functions)"""
    x, p, n = symbols('x p n')
    
    return [
        # ===== INFORMATION THEORY (40 functions) =====
        ('EntropyShannon[p_]', lambda p: -sum(pi*log(pi, 2) for pi in (p.args if hasattr(p, 'args') else p) if pi > 0)),
        ('JointEntropy[p_xy_]', lambda pxy: -sum(sum(pij*log(pij, 2) for pij in row if pij > 0) for row in pxy)),
        ('ConditionalEntropyInfo[p_xy_, p_x_]', lambda pxy, px: sp.Symbol('H(X,Y)') - sp.Symbol('H(X)')),
        ('MutualInformationInfo[H_X_, H_Y_, H_XY_]', lambda HX, HY, HXY: HX + HY - HXY),
        ('RelativeEntropy[p_, q_]', lambda p, q: sum(pi*log(pi/qi, 2) for pi, qi in zip(p, q) if pi > 0)),
        ('KullbackLeiblerDivergence[p_, q_]', lambda p, q: sum(pi*log(pi/qi, 2) for pi, qi in zip(p, q) if pi > 0)),
        ('CrossEntropyInfo[p_, q_]', lambda p, q: -sum(pi*log(qi, 2) for pi, qi in zip(p, q) if pi > 0)),
        ('DifferentialEntropy[f_]', lambda f: -sp.integrate(f*log(f, 2), (sp.Symbol('x'), -sp.oo, sp.oo))),
        ('MaxEntropyPrinciple[constraints_]', lambda c: "max entropy distribution"),
        ('InformationGain[H_before_, H_after_]', lambda Hb, Ha: Hb - Ha),
        ('GiniImpurity[p_]', lambda p: 1 - sum(pi**2 for pi in p)),
        ('ChannelCapacity[C_]', lambda C: C),  # bits/s
        ('ShannonHartleyTheorem[B_, SNR_]', lambda B, SNR: B*log(1 + SNR, 2)),
        ('NyquistRate[f_max_]', lambda fmax: 2*fmax),
        ('SamplingTheorem[f_s_, f_max_]', lambda fs, fmax: fs >= 2*fmax),
        ('SourceCoding[H_, R_]', lambda H, R: R >= H),  # Source coding theorem
        ('ChannelCodingTheorem[R_, C_]', lambda R, C: R < C),
        ('RateDistortionFunction[D_]', lambda D: sp.Symbol('R(D)')),
        ('BinarySymmetricChannel[p_]', lambda p: 1 - sp.Symbol('H')(p)),
        ('BinaryErasureChannel[epsilon_]', lambda eps: 1 - eps),
        ('GaussianChannel[P_, N_]', lambda P, N: sp.Rational(1,2)*log(1 + P/N, 2)),
        ('FanoInequality[Pe_, H_]', lambda Pe, H: sp.Symbol('H(X|Y)') <= sp.Symbol('h')(Pe) + Pe*log(sp.Symbol('|X|') - 1, 2)),
        ('DataProcessingInequality[I_XY_, I_YZ_]', lambda IXY, IYZ: sp.Symbol('I(X;Z)') <= min(IXY, IYZ)),
        ('AsymptoticEquipartitionProperty[n_, H_]', lambda n, H: 2**(-n*H)),
        ('TypicalSet[epsilon_, n_, H_]', lambda eps, n, H: 2**(-n*(H + eps))),
        ('KolmogorovComplexity[K_]', lambda K: K),
        ('AlgorithmicInformation[x_]', lambda x: sp.Symbol('K(x)')),
        ('IncompressibilityTheorem[random_]', lambda r: sp.Symbol('K(x)') >= sp.Symbol('|x|')),
        ('MinimumDescriptionLength[model_, data_]', lambda m, d: sp.Symbol('L(m)') + sp.Symbol('L(d|m)')),
        ('BayesianInformationCriterion[L_, k_, n_]', lambda L, k, n: -2*L + k*log(n)),
        ('AkaikeInformationCriterion[L_, k_]', lambda L, k: -2*L + 2*k),
        ('FisherInformation[score_]', lambda s: sp.Symbol('E')[s**2]),
        ('CramerRaoBound[I_]', lambda I: 1/I),
        ('SufficientStatistic[T_]', lambda T: T),
        ('MinimalSufficientStatistic[T_]', lambda T: T),
        ('ExponentialFamily[theta_]', lambda theta: exp(sp.Symbol('eta')*theta)),
        ('MaximumEntropyDist[constraints_]', lambda c: "distribution"),
        ('InformationGeometry[manifold_]', lambda m: m),
        ('AmariChentsovTensor[metric_]', lambda g: g),
        ('InformationProjection[p_, q_]', lambda p, q: sp.Symbol('KL(p||q)')),
        
        # ===== CODING THEORY (40 functions) =====
        ('HammingDistance[x_, y_]', lambda x, y: sum(xi != yi for xi, yi in zip(str(x), str(y)))),
        ('HammingWeight[x_]', lambda x: sum(int(xi) for xi in str(x))),
        ('MinimumDistance[code_]', lambda c: 1),  # Placeholder
        ('CodeRate[k_, n_]', lambda k, n: k/n),
        ('RedundancyBits[n_, k_]', lambda n, k: n - k),
        ('SpherePackingBound[n_, t_]', lambda n, t: 2**n / sum(sp.binomial(n, i) for i in range(int(t)+1))),
        ('HammingBound[n_, d_]', lambda n, d: 2**n / sum(sp.binomial(n, i) for i in range(int((d-1)/2)+1))),
        ('SingletonBound[n_, k_, d_]', lambda n, k, d: d <= n - k + 1),
        ('PlotkinBound[n_, d_]', lambda n, d: 2**n <= 2*d if 2*d > n else 2**(n-1)),
        ('GilbertVarshamovBound[n_, d_]', lambda n, d: 2**n >= sum(sp.binomial(n-1, i) for i in range(int(d-1)))),
        ('PerfectCode[n_, d_]', lambda n, d: 2**n == sum(sp.binomial(n, i) for i in range(int((d-1)/2)+1))),
        ('HammingCode[m_]', lambda m: (2**m - 1, 2**m - m - 1, 3)),
        ('GolayCode[]', lambda: (23, 12, 7)),
        ('ReedMullerCode[r_, m_]', lambda r, m: (2**m, sum(sp.binomial(m, i) for i in range(int(r)+1)), 2**(m-r))),
        ('BCHCode[n_, t_]', lambda n, t: (n, sp.Symbol('k'), 2*t + 1)),
        ('ReedSolomonCode[n_, k_]', lambda n, k: (n, k, n - k + 1)),
        ('CyclicCode[g_]', lambda g: g),  # Generator polynomial
        ('GeneratorMatrix[G_]', lambda G: G),
        ('ParityCheckMatrix[H_]', lambda H: H),
        ('SyndromeDecoding[s_]', lambda s: s),
        ('ErrorDetectionCapability[d_]', lambda d: d - 1),
        ('ErrorCorrectionCapability[d_]', lambda d: (d - 1)/2),
        ('BurstErrorLength[b_]', lambda b: b),
        ('InterleavingDepth[n_]', lambda n: n),
        ('ConvolutionalCode[constraint_length_]', lambda K: K),
        ('TrellisCode[states_]', lambda s: s),
        ('ViterbiAlgorithm[trellis_]', lambda t: "decoded sequence"),
        ('TurboCode[interleaver_]', lambda i: i),
        ('LDPCCode[parity_check_]', lambda H: H),
        ('BeliefPropagationDecoding[graph_]', lambda g: "decoded"),
        ('SumProductAlgorithm[messages_]', lambda m: m),
        ('PolarCode[N_]', lambda N: N),
        ('SuccessiveCancellation[y_]', lambda y: "decoded"),
        ('FountainCode[k_]', lambda k: k),
        ('RaptorCode[overhead_]', lambda oh: oh),
        ('LTCode[degree_distribution_]', lambda dd: dd),
        ('NetworkCoding[flows_]', lambda f: f),
        ('ReedSolomonErasureCode[n_, k_]', lambda n, k: (n, k)),
        ('MaximumLikelihoodDecoding[y_, code_]', lambda y, c: "ML estimate"),
        ('MaximumAPosteridecoding[y_, prior_]', lambda y, pr: "MAP estimate"),
        
        # ===== CRYPTOGRAPHY (40 functions) =====
        ('CaesarCipher[plaintext_, shift_]', lambda pt, s: ''.join(chr((ord(c) - 65 + int(s)) % 26 + 65) for c in str(pt).upper())),
        ('VigenereCipher[plaintext_, key_]', lambda pt, k: "ciphertext"),
        ('SubstitutionCipher[plaintext_, mapping_]', lambda pt, m: "ciphertext"),
        ('TranspositionCipher[plaintext_, key_]', lambda pt, k: "ciphertext"),
        ('PlayfairCipher[plaintext_, key_]', lambda pt, k: "ciphertext"),
        ('HillCipher[plaintext_, key_matrix_]', lambda pt, K: "ciphertext"),
        ('VernamCipher[plaintext_, key_]', lambda pt, k: "ciphertext"),  # One-time pad
        ('OneTimePad[plaintext_, key_]', lambda pt, k: [pi ^ ki for pi, ki in zip(pt, k)]),
        ('StreamCipher[plaintext_, keystream_]', lambda pt, ks: [pi ^ ksi for pi, ksi in zip(pt, ks)]),
        ('BlockCipher[block_, key_]', lambda b, k: "encrypted block"),
        ('FeistelNetwork[L_, R_, f_, k_]', lambda L, R, f, k: (R, sp.Symbol('L') ^ f(R, k))),
        ('SubstitutionPermutationNetwork[state_, key_]', lambda s, k: "encrypted"),
        ('ConfusionProperty[cipher_]', lambda c: "obscures relationship"),
        ('DiffusionProperty[cipher_]', lambda c: "spreads influence"),
        ('AvalancheEffect[bit_change_]', lambda bc: bc >= 0.5),
        ('ModularExponentiation[base_, exp_, mod_]', lambda b, e, m: pow(int(b), int(e), int(m))),
        ('EuclideanAlgorithm[a_, b_]', lambda a, b: sp.gcd(a, b)),
        ('ExtendedEuclidean[a_, b_]', lambda a, b: (sp.gcd(a, b), sp.Symbol('x'), sp.Symbol('y'))),
        ('ModularInverse[a_, m_]', lambda a, m: pow(int(a), -1, int(m)) if sp.gcd(a, m) == 1 else None),
        ('ChineseRemainderTheorem[remainders_, moduli_]', lambda r, m: sp.Symbol('x')),
        ('FermatLittleTheorem[a_, p_]', lambda a, p: pow(int(a), int(p)-1, int(p)) == 1),
        ('EulerTotientFunction[n_]', lambda n: n - 1 if sp.isprime(n) else sp.Symbol('phi(n)')),
        ('RSAKeyGeneration[p_, q_]', lambda p, q: (p*q, sp.Symbol('e'), sp.Symbol('d'))),
        ('RSAEncryption[m_, e_, n_]', lambda m, e, n: pow(int(m), int(e), int(n))),
        ('RSADecryption[c_, d_, n_]', lambda c, d, n: pow(int(c), int(d), int(n))),
        ('RSASignature[m_, d_, n_]', lambda m, d, n: pow(int(m), int(d), int(n))),
        ('RSAVerification[s_, e_, n_]', lambda s, e, n: pow(int(s), int(e), int(n))),
        ('DiffieHellmanKeyExchange[g_, p_, a_, b_]', lambda g, p, a, b: pow(pow(int(g), int(a), int(p)), int(b), int(p))),
        ('ElGamalEncryption[m_, y_, g_, p_]', lambda m, y, g, p: (sp.Symbol('c1'), sp.Symbol('c2'))),
        ('ElGamalDecryption[c1_, c2_, x_, p_]', lambda c1, c2, x, p: sp.Symbol('m')),
        ('EllipticCurvePoint[x_, y_, a_, b_]', lambda x, y, a, b: y**2 - x**3 - a*x - b),
        ('ECPointAddition[P_, Q_]', lambda P, Q: sp.Symbol('R')),
        ('ECScalarMultiplication[k_, P_]', lambda k, P: sp.Symbol('Q')),
        ('ECDHKeyExchange[P_, a_, b_]', lambda P, a, b: sp.Symbol('shared')),
        ('ECDSASignature[m_, d_, k_]', lambda m, d, k: (sp.Symbol('r'), sp.Symbol('s'))),
        ('ECDSAVerification[m_, r_, s_, Q_]', lambda m, r, s, Q: True),
        ('HashFunction[m_]', lambda m: hash(str(m))),
        ('SHA256[message_]', lambda m: "hash"),
        ('SHA3[message_]', lambda m: "hash"),
        ('HMAC[key_, message_]', lambda k, m: "MAC"),
        
        # ===== DATA COMPRESSION (30 functions) =====
        ('CompressionRatio[original_, compressed_]', lambda orig, comp: orig/comp),
        ('SpaceSavings[original_, compressed_]', lambda orig, comp: (orig - comp)/orig*100),
        ('LosslessCompression[data_]', lambda d: "compressed"),
        ('LossyCompression[data_, quality_]', lambda d, q: "compressed"),
        ('RunLengthEncoding[data_]', lambda d: [(len(list(g)), k) for k, g in sp.Symbol('groupby')(d)]),
        ('HuffmanCoding[frequencies_]', lambda f: "tree"),
        ('HuffmanTree[symbols_, frequencies_]', lambda s, f: "tree"),
        ('OptimalPrefixCode[frequencies_]', lambda f: "Huffman"),
        ('ArithmeticCoding[symbols_, probabilities_]', lambda s, p: "encoded"),
        ('RangeCoding[cumulative_]', lambda cum: "encoded"),
        ('LempelZiv77[data_]', lambda d: "compressed"),
        ('LempelZiv78[data_]', lambda d: "compressed"),
        ('LZW[data_]', lambda d: "compressed"),
        ('DEFLATE[data_]', lambda d: "compressed"),  # LZ77 + Huffman
        ('GZIP[data_]', lambda d: "compressed"),
        ('BZIP2[data_]', lambda d: "compressed"),  # BWT + MTF + Huffman
        ('BurrowsWheelerTransform[text_]', lambda t: "transformed"),
        ('MoveToFrontEncoding[data_]', lambda d: "encoded"),
        ('DeltaEncoding[sequence_]', lambda s: [s[i] - s[i-1] for i in range(1, len(s))]),
        ('DifferentialEncoding[signal_]', lambda s: [s[i] - s[i-1] for i in range(1, len(s))]),
        ('DictionaryCompression[data_]', lambda d: "compressed"),
        ('AdaptiveDictionary[data_]', lambda d: "dictionary"),
        ('StaticDictionary[frequencies_]', lambda f: "dictionary"),
        ('EntropyEncoding[data_]', lambda d: "encoded"),
        ('VariableLengthCode[symbols_]', lambda s: "code"),
        ('UniversalCode[integer_]', lambda n: "code"),
        ('EliasGamma[n_]', lambda n: bin(int(n))[2:]),
        ('EliasDelta[n_]', lambda n: bin(int(n))[2:]),
        ('EliasOmega[n_]', lambda n: bin(int(n))[2:]),
        ('GolombCode[n_, m_]', lambda n, m: (int(n)//int(m), int(n)%int(m))),
    ]


def register():
    """Register all extended29 rules"""
    from .extended_helper import convert_rules
    return convert_rules(get_rules())
